/***************************************************************************

	Namco PuckMan

    driver by Nicola Salmoria and many others

    Games supported:
		* PuckMan
		* Pac-Man Plus
		* Ms. Pac-Man
		* Crush Roller
		* Ponpoko
		* Eyes
		* Mr. TNT
		* Lizard Wizard
		* The Glob
		* Dream Shopper
		* Van Van Car
		* Ali Baba and 40 Thieves
		* Jump Shot
		* Shoot the Bull

	Known issues:
		* mystery items in Ali Baba don't work correctly because of protection

****************************************************************************

	Pac-Man memory map (preliminary)

	0000-3fff ROM
	4000-43ff Video RAM
	4400-47ff Color RAM
	4c00-4fff RAM
	8000-9fff ROM (Ms Pac-Man and Ponpoko only)
	a000-bfff ROM (Ponpoko only)

	memory mapped ports:

	read:
	5000      IN0
	5040      IN1
	5080      DSW 1
	50c0	  DSW 2 (Ponpoko only)
	see the input_ports definition below for details on the input bits

	write:
	4ff0-4fff 8 pairs of two bytes:
	          the first byte contains the sprite image number (bits 2-7), Y flip (bit 0),
			  X flip (bit 1); the second byte the color
	5000      interrupt enable
	5001      sound enable
	5002      ????
	5003      flip screen
	5004      1 player start lamp
	5005      2 players start lamp
	5006      coin lockout
	5007      coin counter
	5040-5044 sound voice 1 accumulator (nibbles) (used by the sound hardware only)
	5045      sound voice 1 waveform (nibble)
	5046-5049 sound voice 2 accumulator (nibbles) (used by the sound hardware only)
	504a      sound voice 2 waveform (nibble)
	504b-504e sound voice 3 accumulator (nibbles) (used by the sound hardware only)
	504f      sound voice 3 waveform (nibble)
	5050-5054 sound voice 1 frequency (nibbles)
	5055      sound voice 1 volume (nibble)
	5056-5059 sound voice 2 frequency (nibbles)
	505a      sound voice 2 volume (nibble)
	505b-505e sound voice 3 frequency (nibbles)
	505f      sound voice 3 volume (nibble)
	5060-506f Sprite coordinates, x/y pairs for 8 sprites
	50c0      Watchdog reset

	I/O ports:
	OUT on port $0 sets the interrupt vector


****************************************************************************

	Make Trax protection description:

	Make Trax has a "Special" chip that it uses for copy protection.
	The following chart shows when reads and writes may occur:

	AAAAAAAA AAAAAAAA
	11111100 00000000  <- address bits
	54321098 76543210
	xxx1xxxx 01xxxxxx - read data bits 4 and 7
	xxx1xxxx 10xxxxxx - read data bits 6 and 7
	xxx1xxxx 11xxxxxx - read data bits 0 through 5

	xxx1xxxx 00xxx100 - write to Special
	xxx1xxxx 00xxx101 - write to Special
	xxx1xxxx 00xxx110 - write to Special
	xxx1xxxx 00xxx111 - write to Special

	In practical terms, it reads from Special when it reads from
	location $5040-$50FF.  Note that these locations overlap our
	inputs and Dip Switches.  Yuk.

	I don't bother trapping the writes right now, because I don't
	know how to interpret them.  However, comparing against Crush
	Roller gives most of the values necessary on the reads.

	Instead of always reading from $5040, $5080, and $50C0, the Make
	Trax programmers chose to read from a wide variety of locations,
	probably to make debugging easier.  To us, it means that for the most
	part we can just assign a specific value to return for each address and
	we'll be OK.  This falls apart for the following addresses:  $50C0, $508E,
	$5090, and $5080.  These addresses should return multiple values.  The other
	ugly thing happening is in the ROMs at $3AE5.  It keeps checking for
	different values of $50C0 and $5080, and weird things happen if it gets
	the wrong values.  The only way I've found around these is to patch the
	ROMs using the same patches Crush Roller uses.  The only thing to watch
	with this is that changing the ROMs will break the beginning checksum.
	That's why we use the rom opcode decode function to do our patches.

	Incidentally, there are extremely few differences between Crush Roller
	and Make Trax.  About 98% of the differences appear to be either unused
	bytes, the name of the game, or code related to the protection.  I've
	only spotted two or three actual differences in the games, and they all
	seem minor.

	If anybody cares, here's a list of disassembled addresses for every
	read and write to the Special chip (not all of the reads are
	specifically for checking the Special bits, some are for checking
	player inputs and Dip Switches):

	Writes: $0084, $012F, $0178, $023C, $0C4C, $1426, $1802, $1817,
		$280C, $2C2E, $2E22, $3205, $3AB7, $3ACC, $3F3D, $3F40,
		$3F4E, $3F5E
	Reads:  $01C8, $01D2, $0260, $030E, $040E, $0416, $046E, $0474,
		$0560, $0568, $05B0, $05B8, $096D, $0972, $0981, $0C27,
		$0C2C, $0F0A, $10B8, $10BE, $111F, $1127, $1156, $115E,
		$11E3, $11E8, $18B7, $18BC, $18CA, $1973, $197A, $1BE7,
		$1C06, $1C9F, $1CAA, $1D79, $213D, $2142, $2389, $238F,
		$2AAE, $2BF4, $2E0A, $39D5, $39DA, $3AE2, $3AEA, $3EE0,
		$3EE9, $3F07, $3F0D

***************************************************************************/

#include "driver.h"
#include "vidhrdw/generic.h"
#include "pacman.h"
#include "pengo.h"

/* PacMAME */
//void machine_init_fastpacman(void);
//int fastpacman_interrupt(void);
//void machine_init_plusfast(void);
//int plusfast_interrupt(void);
static int gamebank;
static int multipac_bankaddress;
void mschamp_decode(void);
/*
static unsigned char *nvram;
static size_t nvram_size;
*/
/* End PacMAME */

// BEGIN Modifications
int multipac_bankaddress;
extern int gfx_bank;
extern void multipac_vh_convert_color_prom(unsigned char *palette, unsigned short *colortable,const unsigned char *color_prom);
extern void gfx_bank_zero(void);
extern void z80_reset(void);
WRITE_HANDLER( superabc_gfxbank_w );
WRITE_HANDLER( multipac_gfxbank_w );
WRITE_HANDLER( multipac_color_w );
static unsigned char *nvram;
static size_t nvram_size;

static void nvram_handler(void *file,int read_or_write)
{
	if (read_or_write)
		osd_fwrite(file,nvram,nvram_size);
	else
	{
		if (file)
			osd_fread(file,nvram,nvram_size);
		else
			memset(nvram,0,nvram_size);
	}
}
// END Modifications

static UINT8 speedcheat = 0;	/* a well known hack allows to make Pac Man run at four times */
								/* his usual speed. When we start the emulation, we check if the */
								/* hack can be applied, and set this flag accordingly. */



/*************************************
 *
 *	Machine init
 *
 *************************************/

/*********************************************************************
SUPER ABC & MULTIPAC
*********************************************************************/
extern void multipac_vh_convert_color_prom(unsigned char *palette, unsigned short *colortable,const unsigned char *color_prom);
WRITE_HANDLER( superabc_gfxbank_w );
WRITE_HANDLER( multipac_gfxbank_w );

MACHINE_INIT( pacman )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* check if the loaded set of ROMs allows the Pac Man speed hack */
	if ((RAM[0x180b] == 0xbe && RAM[0x1ffd] == 0x00) ||
			(RAM[0x180b] == 0x01 && RAM[0x1ffd] == 0xbd))
		speedcheat = 1;
	else
		speedcheat = 0;
}


MACHINE_INIT( mschamp )
{
	data8_t *rom = memory_region(REGION_CPU1) + 0x10000;
	int bankaddr = ((readinputport(3) & 1) * 0x8000);

	cpu_setbank(1,&rom[bankaddr]);
	cpu_setbank(2,&rom[bankaddr+0x4000]);
}

//PacMAME
MACHINE_INIT( fastpacman )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* check if the loaded set of ROMs allows the Pac Man speed hack */
	if ((RAM[0x180b] == 0xbe && RAM[0x1ffd] == 0x00) ||
			(RAM[0x180b] == 0x01 && RAM[0x1ffd] == 0xbd))
		speedcheat = 1;
	else
		speedcheat = 1;
}

MACHINE_INIT( plusfast )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* check if the loaded set of ROMs allows the Pac Man speed hack */
	if ((RAM[0x182d] == 0xbe && RAM[0x1ffd] == 0xff) ||
			(RAM[0x182d] == 0x01 && RAM[0x1ffd] == 0xbe))
		speedcheat = 1;
	else
		speedcheat = 0;
}

/*************************************
 *
 *	Interrupts
 *
 *************************************/

static INTERRUPT_GEN( pacman_interrupt )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* speed up cheat */
	if (speedcheat)
	{
		if (readinputport(4) & 1)	/* check status of the fake dip switch */
		{
			/* activate the cheat */
			RAM[0x180b] = 0x01;
			RAM[0x1ffd] = 0xbd;
		}
		else
		{
			/* remove the cheat */
			RAM[0x180b] = 0xbe;
			RAM[0x1ffd] = 0x00;
		}
	}

	irq0_line_hold();
}


static INTERRUPT_GEN( mspacman_interrupt )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* speed up cheat */
	if (speedcheat)
	{
		if (readinputport(4) & 1)	/* check status of the fake dip switch */
		{
			/* activate the cheat */
			RAM[0x1180b] = 0x01;
			RAM[0x11ffd] = 0xbd;
		}
		else
		{
			/* remove the cheat */
			RAM[0x1180b] = 0xbe;
			RAM[0x11ffd] = 0x00;
		}
	}

	irq0_line_hold();
}

//PacMAME
static INTERRUPT_GEN( fastpacman_interrupt )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* speed up cheat */
	if (speedcheat)
	{
		if (readinputport(4) & 1)	/* check status of the fake dip switch */
		{
			/* activate the cheat */
			RAM[0x180b] = 0x01;
			RAM[0x1ffd] = 0xbd;
		}
		else
		{
			/* remove the cheat */
			RAM[0x180b] = 0xbe;
			RAM[0x1ffd] = 0x00;
		}
	}

	irq0_line_hold();
}

static INTERRUPT_GEN( plusfast_interrupt )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* speed up cheat */
	if (speedcheat)
	{
		if (readinputport(4) & 1)	/* check status of the fake dip switch */
		{
			/* activate the cheat */
			RAM[0x182d] = 0x01;
			RAM[0x1ffd] = 0xbe;
		}
		else
		{
			/* remove the cheat */
			RAM[0x182d] = 0xbe;
			RAM[0x1ffd] = 0xff;
		}
	}

	irq0_line_hold();
}
/* End PacMAME */

/*************************************
 *
 *	LEDs/coin counters
 *
 *************************************/

static WRITE_HANDLER( pacman_leds_w )
{
	set_led_status(offset,data & 1);
}


static WRITE_HANDLER( pacman_coin_counter_w )
{
	coin_counter_w(offset,data & 1);
}


static WRITE_HANDLER( pacman_coin_lockout_global_w )
{
	coin_lockout_global_w(~data & 0x01);
}

/* PacMAME */
static int timerthing = 0x00;
READ_HANDLER( zolatimer_r )
{
	timerthing = timerthing + 0x01;
	if (timerthing == 0xff) timerthing = 0x00;
	return timerthing;
}

static PORT_READ_START( zolapac_readport)
{ 0x00,0x00, zolatimer_r },
PORT_END

/*
static void init_mschamp(void)
{
	mschamp_decode();
}

void mschamp_init_machine(void)
{
	unsigned char *RAM;
	RAM = memory_region(REGION_CPU1);
	gamebank = ((readinputport(3) & 1) * 0x8000 + 0x10000);
	cpu_setbank (1, &RAM[gamebank]);
	cpu_setbank (2, &RAM[gamebank+0x4000]);

}
static int mschamp_interrupt(void)
{
	unsigned char *RAM;
	RAM = memory_region(REGION_CPU1);
	gamebank = ((readinputport(3) & 1) * 0x8000 + 0x10000);
	cpu_setbank (1, &RAM[gamebank]);
	cpu_setbank (2, &RAM[gamebank+0x4000]);
	return interrupt();
}

void mschamp_decode(void)
{
	int i;
	unsigned char *RAM, temp;
	RAM = memory_region(REGION_GFX1);
	for (i = 0; i < 0x0800; i+= 0x01)
	{
		temp = RAM[0x800 + i];		
		RAM[0x800 + i] = RAM[0x1000 + i];
		RAM[0x1000 + i] = temp;
	}
	RAM = memory_region(REGION_GFX2);
	for (i = 0; i < 0x0800; i+= 0x01)
	{
		temp = RAM[0x800 + i];		
		RAM[0x800 + i] = RAM[0x1000 + i];
		RAM[0x1000 + i] = temp;
	}
}

static MEMORY_READ_START( mschamp_readmem )
{ 0x0000, 0x3fff, MRA_BANK1 },
{ 0x4000, 0x47ff, MRA_RAM },
{ 0x4c00, 0x4fff, MRA_RAM },
{ 0x5000, 0x503f, input_port_0_r },
{ 0x5040, 0x507f, input_port_1_r },
{ 0x5080, 0x50bf, input_port_2_r },
{ 0x8000, 0x9fff, MRA_BANK2 },
MEMORY_END
*/
/* End PacMAME */


/*************************************
 *
 *	Ali Baba sound
 *
 *************************************/

static WRITE_HANDLER( alibaba_sound_w )
{
	/* since the sound region in Ali Baba is not contiguous, translate the
	   offset into the 0-0x1f range */
 	if (offset < 0x10)
		pengo_sound_w(offset, data);
	else if (offset < 0x20)
		spriteram_2[offset - 0x10] = data;
	else
		pengo_sound_w(offset - 0x10, data);
}


static READ_HANDLER( alibaba_mystery_1_r )
{
	// The return value determines what the mystery item is.  Each bit corresponds
	// to a question mark

	return rand() & 0x0f;
}


static READ_HANDLER( alibaba_mystery_2_r )
{
	static int mystery = 0;

	// The single bit return value determines when the mystery is lit up.
	// This is certainly wrong

	mystery++;
	return (mystery >> 10) & 1;
}



/*************************************
 *
 *	Make Trax input handlers
 *
 *************************************/

static READ_HANDLER( maketrax_special_port2_r )
{
	int data = input_port_2_r(offset);
	int pc = activecpu_get_previouspc();

	if ((pc == 0x1973) || (pc == 0x2389)) return data | 0x40;

	switch (offset)
	{
		case 0x01:
		case 0x04:
			data |= 0x40; break;
		case 0x05:
			data |= 0xc0; break;
		default:
			data &= 0x3f; break;
	}

	return data;
}


static READ_HANDLER( maketrax_special_port3_r )
{
	int pc = activecpu_get_previouspc();

	if (pc == 0x040e) return 0x20;

	if ((pc == 0x115e) || (pc == 0x3ae2)) return 0x00;

	switch (offset)
	{
		case 0x00:
			return 0x1f;
		case 0x09:
			return 0x30;
		case 0x0c:
			return 0x00;
		default:
			return 0x20;
	}
}



/*************************************
 *
 *	Zola kludge
 *
 *************************************/

static READ_HANDLER( mschamp_kludge_r )
{
	static UINT8 counter;
	return counter++;
}



/*************************************
 *
 *	Main CPU memory handlers
 *
 *************************************/

static MEMORY_READ_START( readmem )
	{ 0x0000, 0x3fff, MRA_ROM },
	{ 0x4000, 0x47ff, MRA_RAM },	/* video and color RAM */
	{ 0x4c00, 0x4fff, MRA_RAM },	/* including sprite codes at 4ff0-4fff */
	{ 0x5000, 0x503f, input_port_0_r },	/* IN0 */
	{ 0x5040, 0x507f, input_port_1_r },	/* IN1 */
	{ 0x5080, 0x50bf, input_port_2_r },	/* DSW1 */
	{ 0x50c0, 0x50ff, input_port_3_r },	/* DSW2 */
	{ 0x8000, 0xbfff, MRA_ROM },	/* Ms. Pac-Man / Ponpoko only */
MEMORY_END


static MEMORY_WRITE_START( writemem )
	{ 0x0000, 0x3fff, MWA_ROM },
	{ 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
	{ 0x4400, 0x47ff, colorram_w, &colorram },
	{ 0x4c00, 0x4fef, MWA_RAM },
	{ 0x4ff0, 0x4fff, MWA_RAM, &spriteram, &spriteram_size },
	{ 0x5000, 0x5000, interrupt_enable_w },
	{ 0x5001, 0x5001, pengo_sound_enable_w },
	{ 0x5002, 0x5002, MWA_NOP },
	{ 0x5003, 0x5003, pengo_flipscreen_w },
 	{ 0x5004, 0x5005, pacman_leds_w },
// 	{ 0x5006, 0x5006, pacman_coin_lockout_global_w },	this breaks many games
 	{ 0x5007, 0x5007, pacman_coin_counter_w },
	{ 0x5040, 0x505f, pengo_sound_w, &pengo_soundregs },
	{ 0x5060, 0x506f, MWA_RAM, &spriteram_2 },
	{ 0x50c0, 0x50c0, watchdog_reset_w },
	{ 0x8000, 0xbfff, MWA_ROM },	/* Ms. Pac-Man / Ponpoko only */
	{ 0xc000, 0xc3ff, videoram_w }, /* mirror address for video ram, */
	{ 0xc400, 0xc7ef, colorram_w }, /* used to display HIGH SCORE and CREDITS */
	{ 0xffff, 0xffff, MWA_NOP },	/* Eyes writes to this location to simplify code */
MEMORY_END


static MEMORY_READ_START( mschamp_readmem )
	{ 0x0000, 0x3fff, MRA_BANK1 },		/* By Sil: Zola/Ms. Champ */
	{ 0x4000, 0x47ff, MRA_RAM },		/* video and color RAM */
	{ 0x4c00, 0x4fff, MRA_RAM },		/* including sprite codes at 4ff0-4fff */
	{ 0x5000, 0x503f, input_port_0_r },	/* IN0 */
	{ 0x5040, 0x507f, input_port_1_r },	/* IN1 */
	{ 0x5080, 0x50bf, input_port_2_r },	/* DSW */
	{ 0x8000, 0x9fff, MRA_BANK2 },		/* By Sil: Zola/Ms. Champ */
MEMORY_END


static MEMORY_READ_START( mspacman_readmem )
	{ 0x0000, 0x3fff, MRA_BANK1 },
	{ 0x4000, 0x47ff, MRA_RAM },	/* video and color RAM */
	{ 0x4c00, 0x4fff, MRA_RAM },	/* including sprite codes at 4ff0-4fff */
	{ 0x5000, 0x503f, input_port_0_r },	/* IN0 */
	{ 0x5040, 0x507f, input_port_1_r },	/* IN1 */
	{ 0x5080, 0x50bf, input_port_2_r },	/* DSW1 */
	{ 0x50c0, 0x50ff, input_port_3_r },	/* DSW2 */
	{ 0x8000, 0xbfff, MRA_BANK1 },
MEMORY_END


static MEMORY_WRITE_START( mspacman_writemem )
	{ 0x0000, 0x3fff, MWA_BANK1 },
	{ 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
	{ 0x4400, 0x47ff, colorram_w, &colorram },
	{ 0x4c00, 0x4fef, MWA_RAM },
	{ 0x4ff0, 0x4fff, MWA_RAM, &spriteram, &spriteram_size },
	{ 0x5000, 0x5000, interrupt_enable_w },
	{ 0x5001, 0x5001, pengo_sound_enable_w },
	{ 0x5002, 0x5002, MWA_NOP },
	{ 0x5003, 0x5003, pengo_flipscreen_w },
 	{ 0x5004, 0x5005, pacman_leds_w },
	{ 0x5006, 0x5006, mspacman_activate_rom },	/* Not actually, just handy */
// 	{ 0x5006, 0x5006, pacman_coin_lockout_global_w },	this breaks many games
 	{ 0x5007, 0x5007, pacman_coin_counter_w },
	{ 0x5040, 0x505f, pengo_sound_w, &pengo_soundregs },
	{ 0x5060, 0x506f, MWA_RAM, &spriteram_2 },
	{ 0x50c0, 0x50c0, watchdog_reset_w },
	{ 0x8000, 0xbfff, MWA_BANK1 },	/* Ms. Pac-Man / Ponpoko only */
	{ 0xc000, 0xc3ff, videoram_w }, /* mirror address for video ram, */
	{ 0xc400, 0xc7ef, colorram_w }, /* used to display HIGH SCORE and CREDITS */
	{ 0xffff, 0xffff, MWA_NOP },	/* Eyes writes to this location to simplify code */
MEMORY_END


static MEMORY_READ_START( alibaba_readmem )
	{ 0x0000, 0x3fff, MRA_ROM },
	{ 0x4000, 0x47ff, MRA_RAM },	/* video and color RAM */
	{ 0x4c00, 0x4fff, MRA_RAM },	/* including sprite codes at 4ef0-4eff */
	{ 0x5000, 0x503f, input_port_0_r },	/* IN0 */
	{ 0x5040, 0x507f, input_port_1_r },	/* IN1 */
	{ 0x5080, 0x50bf, input_port_2_r },	/* DSW1 */
	{ 0x50c0, 0x50c0, alibaba_mystery_1_r },
	{ 0x50c1, 0x50c1, alibaba_mystery_2_r },
	{ 0x8000, 0x8fff, MRA_ROM },
	{ 0x9000, 0x93ff, MRA_RAM },
	{ 0xa000, 0xa7ff, MRA_ROM },
MEMORY_END


static MEMORY_WRITE_START( alibaba_writemem )
	{ 0x0000, 0x3fff, MWA_ROM },
	{ 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
	{ 0x4400, 0x47ff, colorram_w, &colorram },
	{ 0x4ef0, 0x4eff, MWA_RAM, &spriteram, &spriteram_size },
	{ 0x4c00, 0x4fff, MWA_RAM },
	{ 0x5000, 0x5000, watchdog_reset_w },
 	{ 0x5004, 0x5005, pacman_leds_w },
 	{ 0x5006, 0x5006, pacman_coin_lockout_global_w },
 	{ 0x5007, 0x5007, pacman_coin_counter_w },
	{ 0x5040, 0x506f, alibaba_sound_w, &pengo_soundregs },  /* the sound region is not contiguous */
	{ 0x5060, 0x506f, MWA_RAM, &spriteram_2 }, /* actually at 5050-505f, here to point to free RAM */
	{ 0x50c0, 0x50c0, pengo_sound_enable_w },
	{ 0x50c1, 0x50c1, pengo_flipscreen_w },
	{ 0x50c2, 0x50c2, interrupt_enable_w },
	{ 0x8000, 0x8fff, MWA_ROM },
	{ 0x9000, 0x93ff, MWA_RAM },
	{ 0xa000, 0xa7ff, MWA_ROM },
	{ 0xc000, 0xc3ff, videoram_w }, /* mirror address for video ram, */
	{ 0xc400, 0xc7ef, colorram_w }, /* used to display HIGH SCORE and CREDITS */
MEMORY_END


static MEMORY_READ_START( theglobp_readmem )
	{ 0x0000, 0x3fff, MRA_BANK1 },
	{ 0x4000, 0x47ff, MRA_RAM },	/* video and color RAM */
	{ 0x4c00, 0x4fff, MRA_RAM },	/* including sprite codes at 4ff0-4fff */
	{ 0x5000, 0x503f, input_port_0_r },	/* IN0 */
	{ 0x5040, 0x507f, input_port_1_r },	/* IN1 */
	{ 0x5080, 0x50bf, input_port_2_r },	/* DSW1 */
	{ 0x50c0, 0x50ff, input_port_3_r },	/* DSW2 */
MEMORY_END


static MEMORY_READ_START( vanvan_readmem )
	{ 0x0000, 0x3fff, MRA_ROM },
	{ 0x4000, 0x47ff, MRA_RAM },	/* video and color RAM */
	{ 0x4800, 0x4fff, MRA_RAM },	/* including sprite codes at 4ff0-4fff */
	{ 0x5000, 0x5000, input_port_0_r },	/* IN0 */
	{ 0x5040, 0x5040, input_port_1_r },	/* IN1 */
	{ 0x5080, 0x5080, input_port_2_r },	/* DSW1 */
	{ 0x50c0, 0x50c0, input_port_3_r },	/* DSW2 */
	{ 0x8000, 0x8fff, MRA_ROM },
MEMORY_END


static MEMORY_WRITE_START( vanvan_writemem )
	{ 0x0000, 0x3fff, MWA_ROM },
	{ 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
	{ 0x4400, 0x47ff, colorram_w, &colorram },
	{ 0x4800, 0x4fef, MWA_RAM },
	{ 0x4ff0, 0x4fff, MWA_RAM, &spriteram, &spriteram_size },
	{ 0x5000, 0x5000, interrupt_enable_w },
	{ 0x5001, 0x5001, vanvan_bgcolor_w },
	{ 0x5003, 0x5003, pengo_flipscreen_w },
	{ 0x5005, 0x5006, MWA_NOP },	/* always written together with 5001 */
 	{ 0x5007, 0x5007, pacman_coin_counter_w },
	{ 0x5060, 0x506f, MWA_RAM, &spriteram_2 },
	{ 0x5080, 0x5080, MWA_NOP },	/* ??? toggled before reading 5000 */
	{ 0x50c0, 0x50c0, watchdog_reset_w },
	{ 0x8000, 0x8fff, MWA_ROM },
	{ 0xb800, 0xb87f, MWA_NOP },	/* probably a leftover from development: the Sanritsu version */
									/* writes the color lookup table here, while the Karateko version */
									/* writes garbage. */
MEMORY_END



/*************************************
 *
 *	Main CPU port handlers
 *
 *************************************/

static PORT_WRITE_START( writeport )
	{ 0x00, 0x00, interrupt_vector_w },	/* Pac-Man only */
PORT_END


static PORT_WRITE_START( vanvan_writeport )
	{ 0x01, 0x01, SN76496_0_w },
	{ 0x02, 0x02, SN76496_1_w },
PORT_END


static PORT_WRITE_START( dremshpr_writeport )
	{ 0x06, 0x06, AY8910_write_port_0_w },
	{ 0x07, 0x07, AY8910_control_port_0_w },
PORT_END


static PORT_READ_START( theglobp_readport )
	{ 0x00, 0xff, theglobp_decrypt_rom },	/* Switch protection logic */
PORT_END


static PORT_READ_START( mschamp_readport )
	{ 0x00,0x00, mschamp_kludge_r },
PORT_END



/*************************************
 *
 *	Port definitions
 *
 *************************************/

INPUT_PORTS_START( pacman )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_DIPNAME(0x80, 0x80, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(   0x80, DEF_STR( Upright ) )
	PORT_DIPSETTING(   0x00, DEF_STR( Cocktail ) )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "1" )
	PORT_DIPSETTING(    0x04, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x0c, "5" )
	PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x00, "10000" )
	PORT_DIPSETTING(    0x10, "15000" )
	PORT_DIPSETTING(    0x20, "20000" )
	PORT_DIPSETTING(    0x30, "None" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
	PORT_DIPSETTING(    0x40, "Normal" )
	PORT_DIPSETTING(    0x00, "Hard" )
	PORT_DIPNAME( 0x80, 0x80, "Ghost Names" )
	PORT_DIPSETTING(    0x80, "Normal" )
	PORT_DIPSETTING(    0x00, "Alternate" )

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )

	PORT_START	/* FAKE */
	/* This fake input port is used to get the status of the fire button */
	/* and activate the speedup cheat if it is. */
	PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Speedup Cheat", KEYCODE_LCONTROL, JOYCODE_1_BUTTON1 )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x01, DEF_STR( On ) )
INPUT_PORTS_END


/* Ms. Pac-Man input ports are identical to Pac-Man, the only difference is */
/* the missing Ghost Names dip switch. */
INPUT_PORTS_START( mspacman )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "1" )
	PORT_DIPSETTING(    0x04, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x0c, "5" )
	PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x00, "10000" )
	PORT_DIPSETTING(    0x10, "15000" )
	PORT_DIPSETTING(    0x20, "20000" )
	PORT_DIPSETTING(    0x30, "None" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
	PORT_DIPSETTING(    0x40, "Normal" )
	PORT_DIPSETTING(    0x00, "Hard" )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNUSED )

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )

	PORT_START	/* FAKE */
	/* This fake input port is used to get the status of the fire button */
	/* and activate the speedup cheat if it is. */
	PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Speedup Cheat", KEYCODE_LCONTROL, JOYCODE_1_BUTTON1 )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x01, DEF_STR( On ) )
INPUT_PORTS_END


INPUT_PORTS_START( mschamp )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN3 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "1" )
	PORT_DIPSETTING(    0x04, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x0c, "5" )
	PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x00, "10000" )
	PORT_DIPSETTING(    0x10, "15000" )
	PORT_DIPSETTING(    0x20, "20000" )
	PORT_DIPSETTING(    0x30, "None" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
	PORT_DIPSETTING(    0x40, "Normal" )
	PORT_DIPSETTING(    0x00, "Hard" )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNUSED )

	PORT_START	/* DSW 2 */
	PORT_DIPNAME( 0x01, 0x01, "Game" )
	PORT_DIPSETTING(    0x01, "Champion Edition" )
	PORT_DIPSETTING(    0x00, "Super Zola Pac Gal" )
	PORT_DIPNAME( 0x02, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x02, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x04, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x08, DEF_STR( On ) )
INPUT_PORTS_END


INPUT_PORTS_START( maketrax )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Cocktail ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection */
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection */

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x00, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "3" )
	PORT_DIPSETTING(    0x04, "4" )
	PORT_DIPSETTING(    0x08, "5" )
	PORT_DIPSETTING(    0x0c, "6" )
	PORT_DIPNAME( 0x10, 0x10, "First Pattern" )
	PORT_DIPSETTING(    0x10, "Easy" )
	PORT_DIPSETTING(    0x00, "Hard" )
	PORT_DIPNAME( 0x20, 0x20, "Teleport Holes" )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
 	PORT_BIT( 0xc0, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection */

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( mbrush )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Cocktail ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection in Make Trax */
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection in Make Trax */

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "1" )
	PORT_DIPSETTING(    0x04, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x0c, "4" )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x20, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0xc0, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection in Make Trax */

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( paintrlr )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Cocktail ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection in Make Trax */
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection in Make Trax */

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x00, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "3" )
	PORT_DIPSETTING(    0x04, "4" )
	PORT_DIPSETTING(    0x08, "5" )
	PORT_DIPSETTING(    0x0c, "6" )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x20, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0xc0, IP_ACTIVE_HIGH, IPT_UNUSED )  /* Protection in Make Trax */

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( ponpoko )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_HIGH, IPT_JOYSTICK_UP    | IPF_8WAY )
	PORT_BIT( 0x02, IP_ACTIVE_HIGH, IPT_JOYSTICK_LEFT  | IPF_8WAY )
	PORT_BIT( 0x04, IP_ACTIVE_HIGH, IPT_JOYSTICK_RIGHT | IPF_8WAY )
	PORT_BIT( 0x08, IP_ACTIVE_HIGH, IPT_JOYSTICK_DOWN  | IPF_8WAY )
	PORT_BIT( 0x10, IP_ACTIVE_HIGH, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	/* The 2nd player controls are used even in upright mode */
	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_HIGH, IPT_JOYSTICK_UP    | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x02, IP_ACTIVE_HIGH, IPT_JOYSTICK_LEFT  | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x04, IP_ACTIVE_HIGH, IPT_JOYSTICK_RIGHT | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x08, IP_ACTIVE_HIGH, IPT_JOYSTICK_DOWN  | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x10, IP_ACTIVE_HIGH, IPT_BUTTON1 | IPF_PLAYER2 )
	PORT_BIT( 0x20, IP_ACTIVE_HIGH, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_HIGH, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_HIGH, IPT_UNUSED )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x01, "10000" )
	PORT_DIPSETTING(    0x02, "30000" )
	PORT_DIPSETTING(    0x03, "50000" )
	PORT_DIPSETTING(    0x00, "None" )
	PORT_DIPNAME( 0x0c, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, "0" )
	PORT_DIPSETTING(    0x04, "1" )
	PORT_DIPSETTING(    0x08, "2" )
	PORT_DIPSETTING(    0x0c, "3" )
	PORT_DIPNAME( 0x30, 0x20, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "2" )
	PORT_DIPSETTING(    0x10, "3" )
	PORT_DIPSETTING(    0x20, "4" )
	PORT_DIPSETTING(    0x30, "5" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START	/* DSW 2 */
	PORT_DIPNAME( 0x0f, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x04, "A 3/1 B 3/1" )
	PORT_DIPSETTING(    0x0e, "A 3/1 B 1/2" )
	PORT_DIPSETTING(    0x0f, "A 3/1 B 1/4" )
	PORT_DIPSETTING(    0x02, "A 2/1 B 2/1" )
	PORT_DIPSETTING(    0x0d, "A 2/1 B 1/1" )
	PORT_DIPSETTING(    0x07, "A 2/1 B 1/3" )
	PORT_DIPSETTING(    0x0b, "A 2/1 B 1/5" )
	PORT_DIPSETTING(    0x0c, "A 2/1 B 1/6" )
	PORT_DIPSETTING(    0x01, "A 1/1 B 1/1" )
	PORT_DIPSETTING(    0x06, "A 1/1 B 4/5" )
	PORT_DIPSETTING(    0x05, "A 1/1 B 2/3" )
	PORT_DIPSETTING(    0x0a, "A 1/1 B 1/3" )
	PORT_DIPSETTING(    0x08, "A 1/1 B 1/5" )
	PORT_DIPSETTING(    0x09, "A 1/1 B 1/6" )
	PORT_DIPSETTING(    0x03, "A 1/2 B 1/2" )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Unknown ) )  /* Most likely unused */
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x20, DEF_STR( Unknown ) )  /* Most likely unused */
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x00, DEF_STR( Demo_Sounds ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )  /* Most likely unused */
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
INPUT_PORTS_END


INPUT_PORTS_START( eyes )
	PORT_START  /* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_TILT )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN2 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_COCKTAIL )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x03, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x0c, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x04, "4" )
	PORT_DIPSETTING(    0x00, "5" )
	PORT_DIPNAME( 0x30, 0x30, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x30, "50000" )
	PORT_DIPSETTING(    0x20, "75000" )
	PORT_DIPSETTING(    0x10, "100000" )
	PORT_DIPSETTING(    0x00, "125000" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )  /* Not accessed */
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( mrtnt )
	PORT_START  /* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_TILT )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN2 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_COCKTAIL )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x03, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x0c, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x04, "4" )
	PORT_DIPSETTING(    0x00, "5" )
	PORT_DIPNAME( 0x30, 0x30, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x30, "75000" )
	PORT_DIPSETTING(    0x20, "100000" )
	PORT_DIPSETTING(    0x10, "125000" )
	PORT_DIPSETTING(    0x00, "150000" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( lizwiz )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_8WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_8WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_8WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_8WAY )
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_TILT )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN2 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_PLAYER2 )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x03, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x0c, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x04, "4" )
	PORT_DIPSETTING(    0x00, "5" )
	PORT_DIPNAME( 0x30, 0x30, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x30, "75000" )
	PORT_DIPSETTING(    0x20, "100000" )
	PORT_DIPSETTING(    0x10, "125000" )
	PORT_DIPSETTING(    0x00, "150000" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
	PORT_DIPSETTING(    0x40, "Normal" )
	PORT_DIPSETTING(    0x00, "Hard" )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( theglobp )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_UNKNOWN )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_UNKNOWN )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_BUTTON2 | IPF_COCKTAIL )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_COCKTAIL )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_BUTTON2 )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x03, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x03, "3" )
	PORT_DIPSETTING(    0x02, "4" )
	PORT_DIPSETTING(    0x01, "5" )
	PORT_DIPSETTING(    0x00, "6" )
	PORT_DIPNAME( 0x1c, 0x1c, DEF_STR( Difficulty ) )
	PORT_DIPSETTING(    0x1c, "Easiest" )
	PORT_DIPSETTING(    0x18, "Very Easy" )
	PORT_DIPSETTING(    0x14, "Easy" )
	PORT_DIPSETTING(    0x10, "Normal" )
	PORT_DIPSETTING(    0x0c, "Difficult" )
	PORT_DIPSETTING(    0x08, "Very Difficult" )
	PORT_DIPSETTING(    0x04, "Very Hard" )
	PORT_DIPSETTING(    0x00, "Hardest" )
	PORT_DIPNAME( 0x20, 0x00, DEF_STR( Demo_Sounds ) )
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START	/* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( vanvan )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_UNKNOWN )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	/* The 2nd player controls are used even in upright mode */
	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_COCKTAIL )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNKNOWN )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x01, 0x00, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x01, DEF_STR( Cocktail ) )
	PORT_DIPNAME( 0x02, 0x02, DEF_STR( Flip_Screen ) )
	PORT_DIPSETTING(    0x02, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x04, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x08, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x08, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x30, 0x30, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x30, "3" )
	PORT_DIPSETTING(    0x20, "4" )
	PORT_DIPSETTING(    0x10, "5" )
	PORT_DIPSETTING(    0x00, "6" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Coin_A ) )
	PORT_DIPSETTING(    0x00, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x40, DEF_STR( 1C_1C ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Coin_B ) )
	PORT_DIPSETTING(    0x80, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( 1C_3C ) )

	PORT_START	/* DSW 2 */
	PORT_DIPNAME( 0x01, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x01, DEF_STR( On ) )
	PORT_BITX(    0x02, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Invulnerability", KEYCODE_F1, IP_JOY_NONE )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x02, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x04, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x08, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x10, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x20, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x40, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x80, DEF_STR( On ) )
INPUT_PORTS_END


INPUT_PORTS_START( vanvans )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_UNKNOWN )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN2 )

	/* The 2nd player controls are used even in upright mode */
	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_COCKTAIL )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNKNOWN )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x01, 0x00, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x01, DEF_STR( Cocktail ) )
	PORT_DIPNAME( 0x02, 0x02, DEF_STR( Flip_Screen ) )
	PORT_DIPSETTING(    0x02, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x04, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x08, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x08, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x30, 0x30, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x30, "3" )
	PORT_DIPSETTING(    0x20, "4" )
	PORT_DIPSETTING(    0x10, "5" )
	PORT_DIPSETTING(    0x00, "6" )
	PORT_DIPNAME( 0xc0, 0xc0, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x00, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0xc0, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x80, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x40, DEF_STR( 1C_3C ) )

	PORT_START	/* DSW 2 */
	PORT_DIPNAME( 0x01, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x01, DEF_STR( On ) )
	PORT_BITX(    0x02, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Invulnerability", KEYCODE_F1, IP_JOY_NONE )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x02, DEF_STR( On ) )
	PORT_DIPNAME( 0x04, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x04, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x08, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x10, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x20, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x40, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x00, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x80, DEF_STR( On ) )
INPUT_PORTS_END


INPUT_PORTS_START( dremshpr )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_UNKNOWN )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN2 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_COCKTAIL )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNKNOWN )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x01, 0x01, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x01, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )
	PORT_DIPNAME( 0x02, 0x02, DEF_STR( Flip_Screen ) )
	PORT_DIPSETTING(    0x02, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x0c, 0x0c, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x08, "30000" )
	PORT_DIPSETTING(    0x04, "50000" )
	PORT_DIPSETTING(    0x00, "70000" )
	PORT_DIPSETTING(    0x0c, "None" )
	PORT_DIPNAME( 0x30, 0x30, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x30, "3" )
	PORT_DIPSETTING(    0x20, "4" )
	PORT_DIPSETTING(    0x10, "5" )
	PORT_DIPSETTING(    0x00, "6" )
	PORT_DIPNAME( 0xc0, 0xc0, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x00, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0xc0, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x80, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x40, DEF_STR( 1C_3C ) )

	PORT_START	/* DSW 2 */
  //PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Invulnerability", IP_KEY_NONE, IP_JOY_NONE )
  //PORT_DIPSETTING(    0x00, DEF_STR( Off ) )		/* turning this on crashes puts the */
  //PORT_DIPSETTING(    0x01, DEF_STR( On ) )       /* emulated machine in an infinite loop once in a while */
//	PORT_DIPNAME( 0xff, 0x00, DEF_STR( Unused ) )
	PORT_BIT( 0xfe, IP_ACTIVE_LOW, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( alibaba )
	PORT_START	/* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
	PORT_BITX(0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
	PORT_DIPSETTING(0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(0x00, DEF_STR( On ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN2 )

	PORT_START	/* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
	PORT_BIT( 0x10, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_COCKTAIL )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Cabinet ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Upright ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )

	PORT_START	/* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
	PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
	PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
	PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
	PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
	PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
	PORT_DIPSETTING(    0x00, "1" )
	PORT_DIPSETTING(    0x04, "2" )
	PORT_DIPSETTING(    0x08, "3" )
	PORT_DIPSETTING(    0x0c, "5" )
	PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
	PORT_DIPSETTING(    0x00, "10000" )
	PORT_DIPSETTING(    0x10, "15000" )
	PORT_DIPSETTING(    0x20, "20000" )
	PORT_DIPSETTING(    0x30, "None" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
	PORT_DIPSETTING(    0x40, "Normal" )
	PORT_DIPSETTING(    0x00, "Hard" )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
INPUT_PORTS_END


INPUT_PORTS_START( jumpshot )
	PORT_START /* IN0 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_8WAY )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_8WAY )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_8WAY )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_8WAY )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

	PORT_START /* IN1 */
	PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_8WAY | IPF_PLAYER2 )
	PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_8WAY | IPF_PLAYER2 )
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_PLAYER2  )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START /* DSW 1 */
	PORT_DIPNAME( 0x03, 0x01, "Time"  )
//	PORT_DIPSETTING(    0x00,  "2 Minutes"  )
	PORT_DIPSETTING(    0x02,  "2 Minutes" )
	PORT_DIPSETTING(    0x03,  "3 Minutes" )
	PORT_DIPSETTING(    0x01,  "4 Minutes"  )
	PORT_DIPNAME( 0x04, 0x04, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x04, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x08, 0x08, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x08, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Free_Play ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x00, "2 Players Game" )
	PORT_DIPSETTING(    0x20, "1 Credit" )
	PORT_DIPSETTING(    0x00, "2 Credits" )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START /* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END


INPUT_PORTS_START( shootbul )
	PORT_START /* IN0 */
	PORT_ANALOG( 0x0f, 0x0f, IPT_TRACKBALL_X , 50, 25, 0, 0)
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN3 )

	PORT_START /* IN1 */
	PORT_ANALOG( 0x0f, 0x0f, IPT_TRACKBALL_Y | IPF_REVERSE, 50, 25, 0, 0)
	PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
	PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_BUTTON1 )
	PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START1 )
	PORT_DIPNAME( 0x80, 0x80, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x80, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )

	PORT_START /* DSW 1 */
	PORT_DIPNAME( 0x07, 0x07, "Time"  )
	PORT_DIPSETTING(    0x01, "Short")
	PORT_DIPSETTING(    0x07, "Average" )
	PORT_DIPSETTING(    0x03, "Long" )
	PORT_DIPSETTING(    0x05, "Longer" )
	PORT_DIPSETTING(    0x06, "Longest" )
	PORT_DIPNAME( 0x08, 0x08, "Title Page Sounds"  )
	PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x08, DEF_STR( On ) )
	PORT_DIPNAME( 0x10, 0x10, DEF_STR( Unknown ) )
	PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x20, 0x20, DEF_STR( Unknown ))
	PORT_DIPSETTING(    0x20, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_DIPNAME( 0x40, 0x40, DEF_STR( Unknown ))
	PORT_DIPSETTING(    0x40, DEF_STR( Off ) )
	PORT_DIPSETTING(    0x00, DEF_STR( On ) )
	PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNUSED )

	PORT_START /* DSW 2 */
	PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )
INPUT_PORTS_END

/*PacMAME*/
/*
INPUT_PORTS_START( mschamp )
PORT_START
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
PORT_DIPSETTING(    0x00, DEF_STR( On ) )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN3 )

PORT_START
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
PORT_DIPNAME( 0x80, 0x80, DEF_STR( Cabinet ) )
PORT_DIPSETTING(    0x80, DEF_STR( Upright ) )
PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )

PORT_START
PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
PORT_DIPSETTING(    0x00, "1" )
PORT_DIPSETTING(    0x04, "2" )
PORT_DIPSETTING(    0x08, "3" )
PORT_DIPSETTING(    0x0c, "5" )
PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
PORT_DIPSETTING(    0x00, "10000" )
PORT_DIPSETTING(    0x10, "15000" )
PORT_DIPSETTING(    0x20, "20000" )
PORT_DIPSETTING(    0x30, "None" )
PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
PORT_DIPSETTING(    0x40, "Normal" )
PORT_DIPSETTING(    0x00, "Hard" )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNUSED )

PORT_START
PORT_DIPNAME( 0x01, 0x01, DEF_STR( Game ) )
PORT_DIPSETTING(    0x01, "Championship Edition" )
PORT_DIPSETTING(    0x00, "Super Zola Pac Gal" )
INPUT_PORTS_END
*/

INPUT_PORTS_START( zolapac )
PORT_START
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
PORT_DIPSETTING(    0x00, DEF_STR( On ) )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_COIN3 )

PORT_START
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
PORT_DIPNAME( 0x80, 0x80, DEF_STR( Cabinet ) )
PORT_DIPSETTING(    0x80, DEF_STR( Upright ) )
PORT_DIPSETTING(    0x00, DEF_STR( Cocktail ) )

PORT_START
PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
PORT_DIPSETTING(    0x00, "1" )
PORT_DIPSETTING(    0x04, "2" )
PORT_DIPSETTING(    0x08, "3" )
PORT_DIPSETTING(    0x0c, "5" )
PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
PORT_DIPSETTING(    0x00, "10000" )
PORT_DIPSETTING(    0x10, "15000" )
PORT_DIPSETTING(    0x20, "20000" )
PORT_DIPSETTING(    0x30, "None" )
PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
PORT_DIPSETTING(    0x40, "Normal" )
PORT_DIPSETTING(    0x00, "Hard" )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_UNUSED )

PORT_START
PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )

PORT_START
PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Speedup Cheat", KEYCODE_LCONTROL, JOYCODE_1_BUTTON1 )
PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
PORT_DIPSETTING(    0x01, DEF_STR( On ) )
INPUT_PORTS_END
/*End PacMAME*/


/*************************************
 *
 *	Graphics layouts
 *
 *************************************/

static struct GfxLayout tilelayout =
{
	8,8,	/* 8*8 characters */
    256,    /* 256 characters */
    2,  /* 2 bits per pixel */
    { 0, 4 },   /* the two bitplanes for 4 pixels are packed into one byte */
    { 8*8+0, 8*8+1, 8*8+2, 8*8+3, 0, 1, 2, 3 }, /* bits are packed in groups of four */
    { 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
    16*8    /* every char takes 16 bytes */
};


static struct GfxLayout spritelayout =
{
	16,16,	/* 16*16 sprites */
	64,	/* 64 sprites */
	2,	/* 2 bits per pixel */
	{ 0, 4 },	/* the two bitplanes for 4 pixels are packed into one byte */
	{ 8*8, 8*8+1, 8*8+2, 8*8+3, 16*8+0, 16*8+1, 16*8+2, 16*8+3,
			24*8+0, 24*8+1, 24*8+2, 24*8+3, 0, 1, 2, 3 },
	{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8,
			32*8, 33*8, 34*8, 35*8, 36*8, 37*8, 38*8, 39*8 },
	64*8	/* every sprite takes 64 bytes */
};

/*PacMAME*/
/*
static struct GfxDecodeInfo mschampgfxdecodeinfo[] =
{
	{ REGION_GFX1, 0, &tilelayout,   0, 32 },
	{ REGION_GFX2, 4096, &spritelayout, 0, 32 },
	{ -1 }
};
*/
/*End PacMAME*/

static struct GfxDecodeInfo gfxdecodeinfo[] =
{
	{ REGION_GFX1, 0, &tilelayout,   0, 32 },
	{ REGION_GFX2, 0, &spritelayout, 0, 32 },
	{ -1 } /* end of array */
};


static struct GfxDecodeInfo mschampgfxdecodeinfo[] =
{
	{ REGION_GFX1, 0x0000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x1000, &spritelayout, 0, 32 },
	{ -1 } /* end of array */
};



/*************************************
 *
 *	Sound interfaces
 *
 *************************************/

static struct namco_interface namco_interface =
{
	3072000/32,	/* sample rate */
	3,			/* number of voices */
	100,		/* playback volume */
	REGION_SOUND1	/* memory region */
};


static struct SN76496interface sn76496_interface =
{
	2,
	{ 1789750, 1789750 },	/* 1.78975 MHz ? */
	{ 75, 75 }
};


static struct AY8910interface dremshpr_ay8910_interface =
{
	1,	/* 1 chip */
	14318000/8,	/* 1.78975 MHz ??? */
	{ 50 },
	{ 0 },
	{ 0 },
	{ 0 },
	{ 0 }
};



/*************************************
 *
 *	Machine drivers
 *
 *************************************/

static MACHINE_DRIVER_START( pacman )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(readmem,writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(pacman_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(pacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*32)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END


static MACHINE_DRIVER_START( mspacman )

	/* basic machine hardware */
	MDRV_IMPORT_FROM(pacman)

	MDRV_CPU_MODIFY("main")
	MDRV_CPU_MEMORY(mspacman_readmem,mspacman_writemem)
	MDRV_CPU_VBLANK_INT(mspacman_interrupt,1)

	MDRV_MACHINE_INIT(mspacman)
MACHINE_DRIVER_END


static MACHINE_DRIVER_START( mschamp )

	/* basic machine hardware */
	MDRV_IMPORT_FROM(pacman)

	MDRV_CPU_MODIFY("main")
	MDRV_CPU_MEMORY(mschamp_readmem,writemem)
	MDRV_CPU_PORTS(mschamp_readport,writeport)
	MDRV_CPU_VBLANK_INT(irq0_line_hold,1)

	MDRV_MACHINE_INIT(mschamp)

	/* video hardware */
	MDRV_GFXDECODE(mschampgfxdecodeinfo)
MACHINE_DRIVER_END


static MACHINE_DRIVER_START( theglobp )

	/* basic machine hardware */
	MDRV_IMPORT_FROM(pacman)

	MDRV_CPU_MODIFY("main")
	MDRV_CPU_MEMORY(theglobp_readmem,writemem)
	MDRV_CPU_PORTS(theglobp_readport,writeport)

	MDRV_MACHINE_INIT(theglobp)
MACHINE_DRIVER_END


static MACHINE_DRIVER_START( vanvan )

	/* basic machine hardware */
	MDRV_IMPORT_FROM(pacman)

	MDRV_CPU_MODIFY("main")
	MDRV_CPU_MEMORY(vanvan_readmem,vanvan_writemem)
	MDRV_CPU_PORTS(0,vanvan_writeport)
	MDRV_CPU_VBLANK_INT(nmi_line_pulse,1)

	MDRV_MACHINE_INIT(NULL)

	/* video hardware */
	MDRV_VISIBLE_AREA(2*8, 34*8-1, 0*8, 28*8-1)
	MDRV_VIDEO_UPDATE(vanvan)

	/* sound hardware */
	MDRV_SOUND_REPLACE("namco", SN76496, sn76496_interface)
MACHINE_DRIVER_END


static MACHINE_DRIVER_START( dremshpr )

	/* basic machine hardware */
	MDRV_IMPORT_FROM(pacman)

	MDRV_CPU_MODIFY("main")
	MDRV_CPU_PORTS(0,dremshpr_writeport)
	MDRV_CPU_VBLANK_INT(nmi_line_pulse,1)

	MDRV_MACHINE_INIT(NULL)

	/* sound hardware */
	MDRV_SOUND_REPLACE("namco", AY8910, dremshpr_ay8910_interface)
MACHINE_DRIVER_END


static MACHINE_DRIVER_START( alibaba )

	/* basic machine hardware */
	MDRV_IMPORT_FROM(pacman)

	MDRV_CPU_MODIFY("main")
	MDRV_CPU_MEMORY(alibaba_readmem,alibaba_writemem)
	MDRV_CPU_VBLANK_INT(irq0_line_hold,1)

	MDRV_MACHINE_INIT(NULL)
MACHINE_DRIVER_END

/*PacMAME*/
/*
static struct MachineDriver machine_driver_mschamp =
{
	{
		{
			CPU_Z80,
				18432000/6,
				mschamp_readmem,writemem,zolapac_readport,writeport,
				mschamp_interrupt,1
		}
	},
		60, 2500,
		1,
		mschamp_init_machine,

		36*8, 28*8, { 0*8, 36*8-1, 0*8, 28*8-1 },
		mschampgfxdecodeinfo,
		16, 4*32,
		pacman_vh_convert_color_prom,

		VIDEO_TYPE_RASTER | VIDEO_SUPPORTS_DIRTY,
		0,
		pacman_vh_start,
		generic_vh_stop,
		pengo_vh_screenrefresh,

		0,0,0,0,
	{
		{
			SOUND_NAMCO,
				&namco_interface
		}
	}
};
*/

static MACHINE_DRIVER_START( zolapac )

	/* basic machine hardware */
	MDRV_IMPORT_FROM(pacman)

	MDRV_CPU_MODIFY("main")
	MDRV_CPU_MEMORY(mschamp_readmem,writemem)
	MDRV_CPU_PORTS(mschamp_readport,writeport)
	MDRV_CPU_VBLANK_INT(irq0_line_hold,1)

	MDRV_MACHINE_INIT(mschamp)

	/* video hardware */
	MDRV_GFXDECODE(mschampgfxdecodeinfo)
MACHINE_DRIVER_END

/*
static struct MachineDriver machine_driver_zolapac =
{
	{
		{
			CPU_Z80,
				18432000/6,
				readmem,writemem,zolapac_readport,writeport,
				pacman_interrupt,1
		}
	},
		60, 2500,
		1,
		pacman_init_machine,

		36*8, 28*8, { 0*8, 36*8-1, 0*8, 28*8-1 },
		gfxdecodeinfo,
		16, 4*32,
		pacman_vh_convert_color_prom,

		VIDEO_TYPE_RASTER | VIDEO_SUPPORTS_DIRTY,
		0,
		pacman_vh_start,
		generic_vh_stop,
		pengo_vh_screenrefresh,

		0,0,0,0,
	{
		{
			SOUND_NAMCO,
				&namco_interface
		}
	}
};
*/

static MACHINE_DRIVER_START( fastpac )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(readmem,writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(fastpacman_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(fastpacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*32)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END

static MACHINE_DRIVER_START( plusfast )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(readmem,writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(plusfast_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(plusfast)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*32)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END

/*
static struct MachineDriver machine_driver_plusfast =
{
	{
		{
			CPU_Z80,
				18432000/6,
				readmem,writemem,0,writeport,
				plusfast_interrupt,1
		}
	},
		60, 2500,
		1,
		plusfast_init_machine,

		36*8, 28*8, { 0*8, 36*8-1, 0*8, 28*8-1 },
		gfxdecodeinfo,
		16, 4*32,
		pacman_vh_convert_color_prom,

		VIDEO_TYPE_RASTER | VIDEO_SUPPORTS_DIRTY,
		0,
		pacman_vh_start,
		generic_vh_stop,
		pengo_vh_screenrefresh,

		0,0,0,0,
	{
		{
			SOUND_NAMCO,
				&namco_interface
		}
	}
};
*/
/* End PacMAME */

/*************************************
 *
 *	ROM definitions
 *
 *************************************/

ROM_START( puckman )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "namcopac.6e",  0x0000, 0x1000, 0xfee263b3 )
	ROM_LOAD( "namcopac.6f",  0x1000, 0x1000, 0x39d1fc83 )
	ROM_LOAD( "namcopac.6h",  0x2000, 0x1000, 0x02083b03 )
	ROM_LOAD( "namcopac.6j",  0x3000, 0x1000, 0x7a36fe55 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( puckmod )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "namcopac.6e",  0x0000, 0x1000, 0xfee263b3 )
	ROM_LOAD( "namcopac.6f",  0x1000, 0x1000, 0x39d1fc83 )
	ROM_LOAD( "namcopac.6h",  0x2000, 0x1000, 0x02083b03 )
	ROM_LOAD( "npacmod.6j",   0x3000, 0x1000, 0x7d98d5f5 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( puckmana )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
	ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
	ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
	ROM_LOAD( "prg7",         0x3000, 0x0800, 0xb6289b26 )
	ROM_LOAD( "prg8",         0x3800, 0x0800, 0x17a88c13 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "chg1",         0x0000, 0x0800, 0x2066a0b7 )
	ROM_LOAD( "chg2",         0x0800, 0x0800, 0x3591b89d )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( pacman )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
	ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
	ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
	ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( pacmod )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "pacmanh.6e",   0x0000, 0x1000, 0x3b2ec270 )
	ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
	ROM_LOAD( "pacmanh.6h",   0x2000, 0x1000, 0x18811780 )
	ROM_LOAD( "pacmanh.6j",   0x3000, 0x1000, 0x5c96a733 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacmanh.5e",   0x0000, 0x1000, 0x299fb17a )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( hangly )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
	ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
	ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
	ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( hangly2 )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
	ROM_LOAD( "hangly2.6f",   0x1000, 0x0800, 0x5ba228bb )
	ROM_LOAD( "hangly2.6m",   0x1800, 0x0800, 0xbaf5461e )
	ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
	ROM_LOAD( "hangly2.6j",   0x3000, 0x0800, 0x51305374 )
	ROM_LOAD( "hangly2.6p",   0x3800, 0x0800, 0x427c9d4d )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacmanh.5e",   0x0000, 0x1000, 0x299fb17a )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( newpuckx )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xa8ae23c5 )
	ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
	ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0x197443f8 )
	ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x2e64a3ba )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( pacheart )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )     /* 64k for code */
	ROM_LOAD( "pacheart.pg1", 0x0000, 0x0800, 0xd844b679 )
	ROM_LOAD( "pacheart.pg2", 0x0800, 0x0800, 0xb9152a38 )
	ROM_LOAD( "pacheart.pg3", 0x1000, 0x0800, 0x7d177853 )
	ROM_LOAD( "pacheart.pg4", 0x1800, 0x0800, 0x842d6574 )
	ROM_LOAD( "pacheart.pg5", 0x2000, 0x0800, 0x9045a44c )
	ROM_LOAD( "pacheart.pg6", 0x2800, 0x0800, 0x888f3c3e )
	ROM_LOAD( "pacheart.pg7", 0x3000, 0x0800, 0xf5265c10 )
	ROM_LOAD( "pacheart.pg8", 0x3800, 0x0800, 0x1a21a381 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacheart.ch1", 0x0000, 0x0800, 0xc62bbabf )
	ROM_LOAD( "chg2",         0x0800, 0x0800, 0x3591b89d )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacheart.ch3", 0x0000, 0x0800, 0xca8c184c )
	ROM_LOAD( "pacheart.ch4", 0x0800, 0x0800, 0x1b1d9096 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )  /* timing - not used */
ROM_END


ROM_START( piranha )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "pr1.cpu",      0x0000, 0x1000, 0xbc5ad024 )
	ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
	ROM_LOAD( "pr3.cpu",      0x2000, 0x1000, 0x473c379d )
	ROM_LOAD( "pr4.cpu",      0x3000, 0x1000, 0x63fbf895 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pr5.cpu",      0x0000, 0x0800, 0x3fc4030c )
	ROM_LOAD( "pr7.cpu",      0x0800, 0x0800, 0x30b9a010 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pr6.cpu",      0x0000, 0x0800, 0xf3e9c9d5 )
	ROM_LOAD( "pr8.cpu",      0x0800, 0x0800, 0x133d720d )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( pacplus )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
	ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
	ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
	ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x022c35da )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x4de65cdd )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x063dd53a )
	ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0xe271a166 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( mspacman )
	ROM_REGION( 0x20000, REGION_CPU1, 0 )	/* 64k for code+64k for decrypted code */
	ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
	ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
	ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
	ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )
	ROM_LOAD( "u5",           0x8000, 0x0800, 0xf45fbbcd )
	ROM_LOAD( "u6",           0x9000, 0x1000, 0xa90e7000 )
	ROM_LOAD( "u7",           0xb000, 0x1000, 0xc82cd714 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( mspacmab )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
	ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
	ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
	ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
	ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
	ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( mspacmat )
	ROM_REGION( 0x20000, REGION_CPU1, 0 )	/* 64k for code+64k for decrypted code */
	ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
	ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
	ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
	ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )
	ROM_LOAD( "u5",           0x8000, 0x0800, 0xf45fbbcd )
	ROM_LOAD( "u6pacatk",     0x9000, 0x1000, 0xf6d83f4d )
	ROM_LOAD( "u7",           0xb000, 0x1000, 0xc82cd714 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( mspacpls )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
	ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
	ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
	ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
	ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
	ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( pacgal )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
	ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
	ROM_LOAD( "pacman.7fh",   0x2000, 0x1000, 0x513f4d5c )
	ROM_LOAD( "pacman.7hj",   0x3000, 0x1000, 0x70694c8e )
	ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
	ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "pacman.5ef",   0x0000, 0x0800, 0x65a3ee71 )
	ROM_LOAD( "pacman.5hj",   0x0800, 0x0800, 0x50c7477d )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s129.4a",    0x0020, 0x0100, 0x63efb927 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( mschamp )
	ROM_REGION( 0x20000, REGION_CPU1, 0 )
	ROM_LOAD( "pm4.bin", 0x10000, 0x10000, 0x7d6b6303 )	/* banked */

	ROM_REGION( 0x2000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "pm5.bin", 0x0000, 0x0800, 0x7fe6b9e2 )
	ROM_CONTINUE(        0x1000, 0x0800 )
	ROM_CONTINUE(        0x0800, 0x0800 )
	ROM_CONTINUE(        0x1800, 0x0800 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END


ROM_START( crush )
	ROM_REGION( 2*0x10000, REGION_CPU1, 0 )	/* 64k for code + 64k for opcode copy to hack protection */
	ROM_LOAD( "crushkrl.6e",  0x0000, 0x1000, 0xa8dd8f54 )
	ROM_LOAD( "crushkrl.6f",  0x1000, 0x1000, 0x91387299 )
	ROM_LOAD( "crushkrl.6h",  0x2000, 0x1000, 0xd4455f27 )
	ROM_LOAD( "crushkrl.6j",  0x3000, 0x1000, 0xd59fc251 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "maketrax.5e",  0x0000, 0x1000, 0x91bad2da )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "maketrax.5f",  0x0000, 0x1000, 0xaea79f55 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "crush.4a",     0x0020, 0x0100, 0x2bc5d339 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( crush2 )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "tp1",          0x0000, 0x0800, 0xf276592e )
	ROM_LOAD( "tp5a",         0x0800, 0x0800, 0x3d302abe )
	ROM_LOAD( "tp2",          0x1000, 0x0800, 0x25f42e70 )
	ROM_LOAD( "tp6",          0x1800, 0x0800, 0x98279cbe )
	ROM_LOAD( "tp3",          0x2000, 0x0800, 0x8377b4cb )
	ROM_LOAD( "tp7",          0x2800, 0x0800, 0xd8e76c8c )
	ROM_LOAD( "tp4",          0x3000, 0x0800, 0x90b28fa3 )
	ROM_LOAD( "tp8",          0x3800, 0x0800, 0x10854e1b )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "tpa",          0x0000, 0x0800, 0xc7617198 )
	ROM_LOAD( "tpc",          0x0800, 0x0800, 0xe129d76a )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "tpb",          0x0000, 0x0800, 0xd1899f05 )
	ROM_LOAD( "tpd",          0x0800, 0x0800, 0xd35d1caf )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "crush.4a",     0x0020, 0x0100, 0x2bc5d339 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( crush3 )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "unkmol.4e",    0x0000, 0x0800, 0x49150ddf )
	ROM_LOAD( "unkmol.6e",    0x0800, 0x0800, 0x21f47e17 )
	ROM_LOAD( "unkmol.4f",    0x1000, 0x0800, 0x9b6dd592 )
	ROM_LOAD( "unkmol.6f",    0x1800, 0x0800, 0x755c1452 )
	ROM_LOAD( "unkmol.4h",    0x2000, 0x0800, 0xed30a312 )
	ROM_LOAD( "unkmol.6h",    0x2800, 0x0800, 0xfe4bb0eb )
	ROM_LOAD( "unkmol.4j",    0x3000, 0x0800, 0x072b91c9 )
	ROM_LOAD( "unkmol.6j",    0x3800, 0x0800, 0x66fba07d )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "unkmol.5e",    0x0000, 0x0800, 0x338880a0 )
	ROM_LOAD( "unkmol.5h",    0x0800, 0x0800, 0x4ce9c81f )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "unkmol.5f",    0x0000, 0x0800, 0x752e3780 )
	ROM_LOAD( "unkmol.5j",    0x0800, 0x0800, 0x6e00d2ac )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "crush.4a",     0x0020, 0x0100, 0x2bc5d339 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( maketrax )
	ROM_REGION( 2*0x10000, REGION_CPU1, 0 )	/* 64k for code + 64k for opcode copy to hack protection */
	ROM_LOAD( "maketrax.6e",  0x0000, 0x1000, 0x0150fb4a )
	ROM_LOAD( "maketrax.6f",  0x1000, 0x1000, 0x77531691 )
	ROM_LOAD( "maketrax.6h",  0x2000, 0x1000, 0xa2cdc51e )
	ROM_LOAD( "maketrax.6j",  0x3000, 0x1000, 0x0b4b5e0a )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "maketrax.5e",  0x0000, 0x1000, 0x91bad2da )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "maketrax.5f",  0x0000, 0x1000, 0xaea79f55 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "crush.4a",     0x0020, 0x0100, 0x2bc5d339 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( mbrush )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "mbrush.6e",    0x0000, 0x1000, 0x750fbff7 )
	ROM_LOAD( "mbrush.6f",    0x1000, 0x1000, 0x27eb4299 )
	ROM_LOAD( "mbrush.6h",    0x2000, 0x1000, 0xd297108e )
	ROM_LOAD( "mbrush.6j",    0x3000, 0x1000, 0x6fd719d0 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "tpa",          0x0000, 0x0800, 0xc7617198 )
	ROM_LOAD( "mbrush.5h",    0x0800, 0x0800, 0xc15b6967 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "mbrush.5f",    0x0000, 0x0800, 0xd5bc5cb8 )  /* copyright sign was removed */
	ROM_LOAD( "tpd",          0x0800, 0x0800, 0xd35d1caf )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "crush.4a",     0x0020, 0x0100, 0x2bc5d339 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( paintrlr )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "paintrlr.1",   0x0000, 0x0800, 0x556d20b5 )
	ROM_LOAD( "paintrlr.5",   0x0800, 0x0800, 0x4598a965 )
	ROM_LOAD( "paintrlr.2",   0x1000, 0x0800, 0x2da29c81 )
	ROM_LOAD( "paintrlr.6",   0x1800, 0x0800, 0x1f561c54 )
	ROM_LOAD( "paintrlr.3",   0x2000, 0x0800, 0xe695b785 )
	ROM_LOAD( "paintrlr.7",   0x2800, 0x0800, 0x00e6eec0 )
	ROM_LOAD( "paintrlr.4",   0x3000, 0x0800, 0x0fd5884b )
	ROM_LOAD( "paintrlr.8",   0x3800, 0x0800, 0x4900114a )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "tpa",          0x0000, 0x0800, 0xc7617198 )
	ROM_LOAD( "mbrush.5h",    0x0800, 0x0800, 0xc15b6967 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "mbrush.5f",    0x0000, 0x0800, 0xd5bc5cb8 )  /* copyright sign was removed */
	ROM_LOAD( "tpd",          0x0800, 0x0800, 0xd35d1caf )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "crush.4a",     0x0020, 0x0100, 0x2bc5d339 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( ponpoko )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "ppokoj1.bin",  0x0000, 0x1000, 0xffa3c004 )
	ROM_LOAD( "ppokoj2.bin",  0x1000, 0x1000, 0x4a496866 )
	ROM_LOAD( "ppokoj3.bin",  0x2000, 0x1000, 0x17da6ca3 )
	ROM_LOAD( "ppokoj4.bin",  0x3000, 0x1000, 0x9d39a565 )
	ROM_LOAD( "ppoko5.bin",   0x8000, 0x1000, 0x54ca3d7d )
	ROM_LOAD( "ppoko6.bin",   0x9000, 0x1000, 0x3055c7e0 )
	ROM_LOAD( "ppoko7.bin",   0xa000, 0x1000, 0x3cbe47ca )
	ROM_LOAD( "ppokoj8.bin",  0xb000, 0x1000, 0x04b63fc6 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "ppoko9.bin",   0x0000, 0x1000, 0xb73e1a06 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "ppoko10.bin",  0x0000, 0x1000, 0x62069b5d )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( ponpokov )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "ppoko1.bin",   0x0000, 0x1000, 0x49077667 )
	ROM_LOAD( "ppoko2.bin",   0x1000, 0x1000, 0x5101781a )
	ROM_LOAD( "ppoko3.bin",   0x2000, 0x1000, 0xd790ed22 )
	ROM_LOAD( "ppoko4.bin",   0x3000, 0x1000, 0x4e449069 )
	ROM_LOAD( "ppoko5.bin",   0x8000, 0x1000, 0x54ca3d7d )
	ROM_LOAD( "ppoko6.bin",   0x9000, 0x1000, 0x3055c7e0 )
	ROM_LOAD( "ppoko7.bin",   0xa000, 0x1000, 0x3cbe47ca )
	ROM_LOAD( "ppoko8.bin",   0xb000, 0x1000, 0xb39be27d )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "ppoko9.bin",   0x0000, 0x1000, 0xb73e1a06 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "ppoko10.bin",  0x0000, 0x1000, 0x62069b5d )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( eyes )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "d7",           0x0000, 0x1000, 0x3b09ac89 )
	ROM_LOAD( "e7",           0x1000, 0x1000, 0x97096855 )
	ROM_LOAD( "f7",           0x2000, 0x1000, 0x731e294e )
	ROM_LOAD( "h7",           0x3000, 0x1000, 0x22f7a719 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "d5",           0x0000, 0x1000, 0xd6af0030 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "e5",           0x0000, 0x1000, 0xa42b5201 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s129.4a",    0x0020, 0x0100, 0xd8d78829 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( eyes2 )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "g38201.7d",    0x0000, 0x1000, 0x2cda7185 )
	ROM_LOAD( "g38202.7e",    0x1000, 0x1000, 0xb9fe4f59 )
	ROM_LOAD( "g38203.7f",    0x2000, 0x1000, 0xd618ba66 )
	ROM_LOAD( "g38204.7h",    0x3000, 0x1000, 0xcf038276 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "g38205.5d",    0x0000, 0x1000, 0x03b1b4c7 )  /* this one has a (c) sign */

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "e5",           0x0000, 0x1000, 0xa42b5201 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
	ROM_LOAD( "82s129.4a",    0x0020, 0x0100, 0xd8d78829 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( mrtnt )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "tnt.1",        0x0000, 0x1000, 0x0e836586 )
	ROM_LOAD( "tnt.2",        0x1000, 0x1000, 0x779c4c5b )
	ROM_LOAD( "tnt.3",        0x2000, 0x1000, 0xad6fc688 )
	ROM_LOAD( "tnt.4",        0x3000, 0x1000, 0xd77557b3 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "tnt.5",        0x0000, 0x1000, 0x3038cc0e )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "tnt.6",        0x0000, 0x1000, 0x97634d8b )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "mrtnt08.bin",  0x0000, 0x0020, 0x00000000 )
	ROM_LOAD( "mrtnt04.bin",  0x0020, 0x0100, 0x00000000 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m"  ,  0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( lizwiz )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "6e.cpu",       0x0000, 0x1000, 0x32bc1990 )
	ROM_LOAD( "6f.cpu",       0x1000, 0x1000, 0xef24b414 )
	ROM_LOAD( "6h.cpu",       0x2000, 0x1000, 0x30bed83d )
	ROM_LOAD( "6j.cpu",       0x3000, 0x1000, 0xdd09baeb )
	ROM_LOAD( "wiza",         0x8000, 0x1000, 0xf6dea3a6 )
	ROM_LOAD( "wizb",         0x9000, 0x1000, 0xf27fb5a8 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e.cpu",       0x0000, 0x1000, 0x45059e73 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "5f.cpu",       0x0000, 0x1000, 0xd2469717 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "7f.cpu",       0x0000, 0x0020, 0x7549a947 )
	ROM_LOAD( "4a.cpu",       0x0020, 0x0100, 0x5fdca536 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m"  ,  0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( theglobp )
	ROM_REGION( 0x20000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "glob.u2",      0x0000, 0x2000, 0x829d0bea )
	ROM_LOAD( "glob.u3",      0x2000, 0x2000, 0x31de6628 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "glob.5e",      0x0000, 0x1000, 0x53688260 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "glob.5f",      0x0000, 0x1000, 0x051f59c7 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "glob.7f",      0x0000, 0x0020, 0x1f617527 )
	ROM_LOAD( "glob.4a",      0x0020, 0x0100, 0x28faa769 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m"  ,  0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( beastf )
	ROM_REGION( 0x20000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "bf-u2.bin",    0x0000, 0x2000, 0x3afc517b )
	ROM_LOAD( "bf-u3.bin",    0x2000, 0x2000, 0x8dbd76d0 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "beastf.5e",    0x0000, 0x1000, 0x5654dc34 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "beastf.5f",    0x0000, 0x1000, 0x1b30ca61 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "glob.7f",      0x0000, 0x0020, 0x1f617527 )
	ROM_LOAD( "glob.4a",      0x0020, 0x0100, 0x28faa769 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m"  ,  0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( vanvan )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "van-1.50",     0x0000, 0x1000, 0xcf1b2df0 )
	ROM_LOAD( "van-2.51",     0x1000, 0x1000, 0xdf58e1cb )
	ROM_LOAD( "van-3.52",     0x2000, 0x1000, 0x15571e24 )
	ROM_LOAD( "van-4.53",     0x3000, 0x1000, 0xb724cbe0 )
	ROM_LOAD( "van-5.39",     0x8000, 0x1000, 0xdb67414c )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "van-20.18",    0x0000, 0x1000, 0x60efbe66 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "van-21.19",    0x0000, 0x1000, 0x5dd53723 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "6331-1.6",     0x0000, 0x0020, 0xce1d9503 )
	ROM_LOAD( "6301-1.37",    0x0020, 0x0100, 0x4b803d9f )
ROM_END

ROM_START( vanvank )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "van1.bin",	  0x0000, 0x1000, 0x00f48295 )
	ROM_LOAD( "van-2.51",     0x1000, 0x1000, 0xdf58e1cb )
	ROM_LOAD( "van-3.52",     0x2000, 0x1000, 0x15571e24 )
	ROM_LOAD( "van4.bin",     0x3000, 0x1000, 0xf8b37ed5 )
	ROM_LOAD( "van5.bin",     0x8000, 0x1000, 0xb8c1e089 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "van-20.18",    0x0000, 0x1000, 0x60efbe66 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "van-21.19",    0x0000, 0x1000, 0x5dd53723 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "6331-1.6",     0x0000, 0x0020, 0xce1d9503 )
	ROM_LOAD( "6301-1.37",    0x0020, 0x0100, 0x4b803d9f )
ROM_END

ROM_START( dremshpr )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "red_1.50",	  0x0000, 0x1000, 0x830c6361 )
	ROM_LOAD( "red_2.51",     0x1000, 0x1000, 0xd22551cc )
	ROM_LOAD( "red_3.52",     0x2000, 0x1000, 0x0713a34a )
	ROM_LOAD( "red_4.53",     0x3000, 0x1000, 0xf38bcaaa )
	ROM_LOAD( "red_5.39",     0x8000, 0x1000, 0x6a382267 )
	ROM_LOAD( "red_6.40",     0x9000, 0x1000, 0x4cf8b121 )
	ROM_LOAD( "red_7.41",     0xa000, 0x1000, 0xbd4fc4ba )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "red-20.18",    0x0000, 0x1000, 0x2d6698dc )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "red-21.19",    0x0000, 0x1000, 0x38c9ce9b )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "6331-1.6",     0x0000, 0x0020, 0xce1d9503 )
	ROM_LOAD( "6301-1.37",    0x0020, 0x0100, 0x39d6fb5c )
ROM_END


ROM_START( alibaba )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "6e",           0x0000, 0x1000, 0x38d701aa )
	ROM_LOAD( "6f",           0x1000, 0x1000, 0x3d0e35f3 )
	ROM_LOAD( "6h",           0x2000, 0x1000, 0x823bee89 )
	ROM_LOAD( "6k",           0x3000, 0x1000, 0x474d032f )
	ROM_LOAD( "6l",           0x8000, 0x1000, 0x5ab315c1 )
	ROM_LOAD( "6m",           0xa000, 0x0800, 0x438d0357 )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e",           0x0000, 0x0800, 0x85bcb8f8 )
	ROM_LOAD( "5h",           0x0800, 0x0800, 0x38e50862 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "5f",           0x0000, 0x0800, 0xb5715c86 )
	ROM_LOAD( "5k",           0x0800, 0x0800, 0x713086b3 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "alibaba.7f",   0x0000, 0x0020, 0x00000000 )  /* missing */
	ROM_LOAD( "alibaba.4a",   0x0020, 0x0100, 0x00000000 )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( jumpshot )
	ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
	ROM_LOAD( "6e",           0x0000, 0x1000, 0xf00def9a )
	ROM_LOAD( "6f",           0x1000, 0x1000, 0xf70deae2 )
	ROM_LOAD( "6h",           0x2000, 0x1000, 0x894d6f68 )
	ROM_LOAD( "6j",           0x3000, 0x1000, 0xf15a108a )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "5e",           0x0000, 0x1000, 0xd9fa90f5 )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "5f",           0x0000, 0x1000, 0x2ec711c1 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "prom.7f",      0x0000, 0x0020, 0x872b42f3 )
	ROM_LOAD( "prom.4a",      0x0020, 0x0100, 0x0399f39f )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


ROM_START( shootbul )
	ROM_REGION( 0x10000, REGION_CPU1, 0 ) /* 64k for code */
	ROM_LOAD( "sb6e.cpu",     0x0000, 0x1000, 0x25daa5e9 )
	ROM_LOAD( "sb6f.cpu",     0x1000, 0x1000, 0x92144044 )
	ROM_LOAD( "sb6h.cpu",     0x2000, 0x1000, 0x43b7f99d )
	ROM_LOAD( "sb6j.cpu",     0x3000, 0x1000, 0xbc4d3bbf )

	ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
	ROM_LOAD( "sb5e.cpu",     0x0000, 0x1000, 0x07c6c5aa )

	ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
	ROM_LOAD( "sb5f.cpu",     0x0000, 0x1000, 0xeaec6837 )

	ROM_REGION( 0x0120, REGION_PROMS, 0 )
	ROM_LOAD( "7f.rom",       0x0000, 0x0020, 0xec578b98 )
	ROM_LOAD( "4a.rom",       0x0020, 0x0100, 0x81a6b30f )

	ROM_REGION( 0x0200, REGION_SOUND1, 0 ) /* sound PROMs */
	ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
	ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 ) /* timing - not used */
ROM_END

/* PacMAME Games */
ROM_START( npacmod )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "namcopac.6e",  0x0000, 0x1000, 0xfee263b3 )
ROM_LOAD( "namcopac.6f",  0x1000, 0x1000, 0x39d1fc83 )
ROM_LOAD( "namcopac.6h",  0x2000, 0x1000, 0x02083b03 )
ROM_LOAD( "npacmod.6j",   0x3000, 0x1000, 0x7d98d5f5 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanjp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "prg7",         0x3000, 0x0800, 0xb6289b26 )
ROM_LOAD( "prg8",         0x3800, 0x0800, 0x17a88c13 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "chg1",         0x0000, 0x0800, 0x2066a0b7 )
ROM_LOAD( "chg2",         0x0800, 0x0800, 0x3591b89d )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanm )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( cookiem )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0x3c95b06b )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x527859fb )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x66e397fd )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xa96761a6 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xffb4de27 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x1b5875ac )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

/*
ROM_START( mschamp )
ROM_REGION( 0x20000, REGION_CPU1,0)
ROM_LOAD( "pm4.bin", 0x10000, 0x10000, 0x7d6b6303 )

ROM_REGION( 0x2000, REGION_GFX1, 0 )
ROM_LOAD( "pm5.bin", 0x0000, 0x2000, 0x7fe6b9e2 )

ROM_REGION( 0x2000, REGION_GFX2, 0 )
ROM_LOAD( "pm5.bin", 0x0000, 0x2000, 0x7fe6b9e2 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END
*/

ROM_START( hanglyb )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "01",           0x0000, 0x0800, 0x9d027c4a )
ROM_LOAD( "05",           0x0800, 0x0800, 0x194c7189 )
ROM_LOAD( "02",           0x1000, 0x0800, 0x5ba228bb )
ROM_LOAD( "06",           0x1800, 0x0800, 0x70810ccf )
ROM_LOAD( "03",           0x2000, 0x0800, 0x08419c4a )
ROM_LOAD( "07",           0x2800, 0x0800, 0x3f250c58 )
ROM_LOAD( "04",           0x3000, 0x0800, 0x603b70b7 )
ROM_LOAD( "08",           0x3800, 0x0800, 0x71293b1f )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "09",           0x0000, 0x0800, 0xc62bbabf )
ROM_LOAD( "11",           0x0800, 0x0800, 0x777c70d3 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "10",         0x0000, 0x0800, 0xca8c184c )
ROM_LOAD( "12",           0x0800, 0x0800, 0x7dc75a81 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( sueworld )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x677adea7 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x6c975345 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x2c809842 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacjr1 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xfcfb21a3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x9f3c32d4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xc2310808 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacjr2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xcc26c905 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x9f3c32d4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xc2310808 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacjr3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xd4d8cb9b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x9f3c32d4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xc2310808 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacjr4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x42ca024b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x9f3c32d4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xc2310808 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( zolaslow )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0x2b9e9433 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x694450a3 )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xb5b30242 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x092292f1 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x52209eee )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( caterpil )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xb53c5650 )
ROM_LOAD( "puckman.6f",   0x1000, 0x1000, 0x53845efb )
ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0xa474357e )
ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x5df0ea3b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5e",    0x0000, 0x1000, 0x0387907d )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5f",    0x0000, 0x1000, 0xaa450714 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmulti )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacmulti.6e",    0x0000, 0x1000, 0xcfb721a8 )
ROM_LOAD( "pacmulti.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacmulti.6h",    0x2000, 0x1000, 0xe55d5230 )
ROM_LOAD( "pacmulti.6j",    0x3000, 0x1000, 0x29ba4b53 )
ROM_LOAD( "pacmulti.8",     0x8000, 0x1000, 0x5a4b59dc )
ROM_LOAD( "pacmulti.9",     0x9000, 0x1000, 0x3b939b12 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacmulti.5e",    0x0000, 0x1000, 0x46cdda4a )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacmulti.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacstrm )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x720dc3ee )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xb2940b89 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x5c65865f )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mshangly )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( chtmsatk )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( fstmsatk )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0xa8d6c227 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( zolamaze )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( zolatunl )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0x2b9e9433 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0c8631d1 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x694450a3 )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xb5b30242 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x52209eee )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( ultra2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x83da903e )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msatk2ad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "mspacatk.1",   0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "mspacatk.3",   0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "mspacatk.4",   0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0x5a72a7a6 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0xbd0c3e94 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xdc70ed2e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x643d5523 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msatkad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xdc70ed2e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x643d5523 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( zolaad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )

ROM_LOAD( "1.bin",   0x0000, 0x4000, 0x1aa2f312 )
ROM_LOAD( "2.bin",   0x8000, 0x2000, 0x420ff603 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xdc70ed2e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x643d5523 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( zolavect )
ROM_REGION( 0x10000, REGION_CPU1, 0 )

ROM_LOAD( "1.bin",   0x0000, 0x4000, 0x1aa2f312 )
ROM_LOAD( "2.bin",   0x8000, 0x2000, 0x420ff603 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x56560aef )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x57cb31e3 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( zolaatk )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0x2b9e9433 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0c8631d1 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x694450a3 )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xb5b30242 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x24cb5d9b )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xe0c03b4e )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( hanglyad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x93fd3682 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xc2310808 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( vhangly )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x022d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( xensad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x64a10b6c )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x315f7131 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x157a6c0f )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vectxens )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x64a10b6c )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x022d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( punleash )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xeddb20e7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x57fe8b4d )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x94c63bb1 )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xfe7734d5 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xef155ffb )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x70d15899 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x713a4ba4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x98d3d364 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x357c2523 )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x5ff3b85a )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fpnleash )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xeddb20e7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x57fe8b4d )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x94c63bb1 )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xfe7734d5 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xef155ffb )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x70d15899 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "flat.5e",           0x0000, 0x1000, 0x1f69d1d0 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x98d3d364 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x357c2523 )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x5ff3b85a )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( sumelton )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xc1fe2c7d )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x10894f38 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xdebef8a9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacweird )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0x92fac825 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0x11077f3e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x7c213ab4 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msnes6m2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xb8f8edda )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x5d7888bc )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x3abb116b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x6a4db98d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msnes6m3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x20833779 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xa047e68f )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x3abb116b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x6a4db98d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msnes6m4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x23c0db46 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xdaafdb39 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x22e5b64e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x6a4db98d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msnes4a )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xd16b9605 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xe6b58707 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x22e5b64e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x6a4db98d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmn6m2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0xaadee235 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0x60ada878 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msmini )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x7787a3f6 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x98ab3c44 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( msminia )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x7787a3f6 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x98ab3c44 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspc6mad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xb556fe78 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x4dbe1939 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x92ea9860 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x713cc17e )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( msvctr6m )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xd0bd89d9 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xc5662407 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x64b35a5a )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x57cb31e3 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pmad6m )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0xd6a0d7f2 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0x93fd3682 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0xc2310808 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( snakeyes )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0x7e506e78 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0x46cdda4a )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( tbone )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0xe7377f2d )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0xafd44c48 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vectr6m )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0xde753fd9 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0xf7be09ba )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vectr6tb )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0xbdbbb207 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0xf7be09ba )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( snowpac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x4f9e1d02 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x2cc60a94 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x41ba87bd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msultra )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x36b5e9fd )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xcf609615 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x7cd365 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x508af5b0 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xdb5a4630 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msnes63 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xf4ea7d2c )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x1ea43bba )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xd3a28769 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x6a4db98d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msnes62 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x52120bc9 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x8fd36e02 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xd3a28769 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x6a4db98d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msnes6m )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xbc091971 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x23ebe9a1 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x3abb116b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x6a4db98d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( ms2600 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xc2a293df )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x4852db05 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xf32b076f )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x6b07b97a )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xf844608b )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mspacren )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xefe80423 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xf513ae17 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xce3b842b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x31b134a0 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x22f61d1 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x7b6aa3c9 )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x64fe1dbf )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( baby2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x260b87b8 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x6f7d8d57 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xb6d77a1e )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( baby3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x6c0e22c8 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x22174b65 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xb6d77a1e )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( baby4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xf8cd7ebb )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x22174b65 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xb6d77a1e )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msbaby )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x17aea21d )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xc08aa7f )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xe28aaa1f )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xe764738d )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x3d316ad0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msbaby1 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x17aea21d )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xc08aa7f )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xe28aaa1f )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xe764738d )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x3d316ad0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacman6 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x581ae70b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0xde753fd9 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0x46cdda4a )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( ultrapac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman6.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman6.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman6.6h",    0x2000, 0x1000, 0xaee79ea1 )
ROM_LOAD( "pacman6.6j",    0x3000, 0x1000, 0x84c6c23b )
ROM_LOAD( "pacmaps.rom",   0x8000, 0x1000, 0xbe212894 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5e",    0x0000, 0x1000, 0xd46efbcf )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman6.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( crazypac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x9f754a51 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( dizzy )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xbf05bf15 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xf999a8d9 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x7cb7b9b1 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msrumble )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x15b02313 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xe2473a86 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x17abb097 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mrpacman )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x65c0526c )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x4078d9b2 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xc1637f1c )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x83e6a3a9 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x45dadf0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vcrunchy )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xc21c6c52 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacbell )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x4b428215 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacelec )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x9e5c763d )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacjail )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xcc31f185 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacshuf )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x55cc4d87 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpspeed )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xe7946d1 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacms1 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xc5da3887 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacms2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xb0a107ea )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacms3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xc7d8325a )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacms4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x7b63ff9b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacmsa1 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x7163b97 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacmsa2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x15e83c24 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacmsa3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x5762f9cf )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vpacmsa4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xd9c2a669 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x35f0597 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x306b24f0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vecbaby )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "babypac.6j",    0x3000, 0x1000, 0x4e4ac798 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vecbaby2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "babypac2.6j",    0x3000, 0x1000, 0x742453ba )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vecbaby3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "babypac3.6j",    0x3000, 0x1000, 0xb82f2b73 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( puckren )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xf5ecf67 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x70ea1149 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xc6bb6ea6 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xdc90f1dc )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xa502e25f )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x18d44098 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x7b6aa3c9 )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x64fe1dbf )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( puckrenc )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xf5ecf67 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x8a75a7e2 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xc6bb6ea6 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xdc90f1dc )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xa502e25f )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x18d44098 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x7b6aa3c9 )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x64fe1dbf )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mspacatb )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5b",   0x8000, 0x1000, 0x5a72a7a6 )
ROM_LOAD( "mspacatk.6b",   0x9000, 0x1000, 0x7a8c4dfe )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( zolapac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )

ROM_LOAD( "1.bin",   0x0000, 0x4000, 0x1aa2f312 )
ROM_LOAD( "2.bin",   0x8000, 0x2000, 0x420ff603 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacms1 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xc5da3887 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacms2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xb0a107ea )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacms3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xc7d8325a )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacms4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x7b63ff9b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmsa1 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x7163b97 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmsa2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x15e83c24 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmsa3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x5762f9cf )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmsa4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xd9c2a669 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7394868b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xce1a3264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fastplus )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
/*	ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 ) */
ROM_LOAD( "boot2",   	  0x1000, 0x1000, 0xa8d6c227 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacbaby )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "babypac.6j",    0x3000, 0x1000, 0x4e4ac798 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacbaby2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "babypac2.6j",  0x3000, 0x1000, 0x742453ba )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacbaby3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "babypac3.6j",    0x3000, 0x1000, 0xb82f2b73 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( plusfast )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",    0x0000, 0x1000, 0x22c35da )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",    0x0000, 0x1000, 0x4de65cdd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "pacplus.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pmanaltm )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xfee263b3 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0xc5ec2352 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0x2083b03 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x57a07f6e )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( faststrm )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xa8d6c227 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xb2940b89 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x5c65865f )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mshearts )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xb5cf081 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vectplus )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",    0x0000, 0x1000, 0x330abcc6 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",    0x0000, 0x1000, 0x61da12cf )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "pacplus.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacbell )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x4b428215 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacelec )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x9e5c763d )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacinvis )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x1b4e96dc )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacjail )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xcc31f185 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacshuf )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x55cc4d87 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacspeed )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xe7946d1 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pcrunchy )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xc21c6c52 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacm255 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xd3640977 )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacn255 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "namcopac.6e",  0x0000, 0x1000, 0xfee263b3 )
ROM_LOAD( "namcopac.6f",  0x1000, 0x1000, 0x39d1fc83 )
ROM_LOAD( "namcopac.6h",  0x2000, 0x1000, 0x6dac9e8f )
ROM_LOAD( "namcopac.6j",  0x3000, 0x1000, 0x7a36fe55 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( hearts )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x5be0a4e4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacplusc )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacplusc.rom",   0x0000, 0x4000, 0xa0d25591 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",    0x0000, 0x1000, 0x22c35da )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",    0x0000, 0x1000, 0x4de65cdd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "pacplus.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pmad00 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xdd2eb43d )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x5f1aea94 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x315f7131 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x157a6c0f )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msf1pac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xe255428b )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xc7183fd2 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xa355e061 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( eltonpac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x9c093e5b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xc2825641 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( xensrev )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x64a10b6c )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x40e3522d )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pengman )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x37e7f3ba )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x2f5f5e6 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x5cd3db2e )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( europac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xa4e82007 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x208da4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xd36dabf )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pac2600 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x0a27d11d )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xab23fe0c )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x87d8f5ce )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x91bd671f )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( roboman )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xa909673e )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xe386213e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xd8db5788 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pmad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xdd2eb43d )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0xff3f4866 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x5b794d59 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x59a9362c )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pmada )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xdd2eb43d )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x24e43519 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x9f3c32d4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xc2310808 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmini )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x5e04f9c5 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x5520b393 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x30e6024c )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmini2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x092dd5fa )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x130eff7b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x30e6024c )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacman3d )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x96364259 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x959e930e )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xaa203d45 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xd1830540 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vectrpac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x936d0475 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x22d0686 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( brakman )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x2d2af2e5 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xa004abe7 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x30ed6264 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fastpac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x720dc3ee )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( chtpac )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x72cea888 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fastpop )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x720dc3ee )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xaf2b610b )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( chtpop )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacman.6e",    0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",    0x1000, 0x1000, 0x72cea888 )
ROM_LOAD( "pacman.6h",    0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",    0x3000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xaf2b610b )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pmanalt )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "namcopac.6e",    0x0000, 0x1000, 0xfee263b3 )
ROM_LOAD( "namcopac.6f",    0x1000, 0x1000, 0xc5ec2352 )
ROM_LOAD( "namcopac.6h",    0x2000, 0x1000, 0x02083b03 )
ROM_LOAD( "namcopac.6j",    0x3000, 0x1000, 0x57a07f6e )
/*
ROM_REGION_DISPOSE(0x2000)
ROM_LOAD( "namcopac.5e",    0x0000, 0x1000, 0x0c944964 )
ROM_LOAD( "namcopac.5f",    0x1000, 0x1000, 0x958fedf9 )
*/
ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "namcopac.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "namcopac.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fasthang )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x1b1014bc )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "hangly.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "hangly.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( chthang )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "hangly.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "hangly.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mazeman )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xa8ae23c5 )
ROM_LOAD( "puckman.6f",   0x1000, 0x1000, 0x77e44852 )
ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0x6d16a528 )
ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x3c3ff208 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5e",    0x0000, 0x1000, 0x38870a32 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5f",    0x0000, 0x1000, 0x1fd18b80 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fastpuck )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xa8ae23c5 )
ROM_LOAD( "puckman.6f",   0x1000, 0x1000, 0x720dc3ee )
ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0x197443f8 )
ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x2e64a3ba )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( chtpuck )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xa8ae23c5 )
ROM_LOAD( "puckman.6f",   0x1000, 0x1000, 0x72cea888 )
ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0x197443f8 )
ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x2e64a3ba )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( puckman2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xa8ae23c5 )
ROM_LOAD( "puckman.6f",   0x1000, 0x1000, 0x720dc3ee )
ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0x197443f8 )
ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x0e3bdbde )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fstpman2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xa8ae23c5 )
ROM_LOAD( "puckman.6f",   0x1000, 0x1000, 0x720dc3ee )
ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0x197443f8 )
ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x0e3bdbde )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( chtpman2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "puckman.6e",   0x0000, 0x1000, 0xa8ae23c5 )
ROM_LOAD( "puckman.6f",   0x1000, 0x1000, 0x72cea888 )
ROM_LOAD( "puckman.6h",   0x2000, 0x1000, 0x197443f8 )
ROM_LOAD( "puckman.6j",   0x3000, 0x1000, 0x0e3bdbde )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5e",    0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "puckman.5f",    0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pplusad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",    0x0000, 0x1000, 0xce684d85 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",    0x0000, 0x1000, 0xb86d002 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "pacplus.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mssilad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x79d241dd )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x2e2fcde0 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x22c6d19a )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xdc70ed2e )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x643d5523 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mspac6m )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "1.cpu",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "2.cpu",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "3.cpu",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "4.cpu",        0x3000, 0x1000, 0xe21c81ff )
ROM_LOAD( "52.cpu",       0x8000, 0x1000, 0xd0bd89d9 )
ROM_LOAD( "62.cpu",       0x9000, 0x1000, 0xc5662407 )

ROM_REGION( 0x2000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "7.cpu",           0x0000, 0x0800, 0x2850148a )
ROM_LOAD( "9.cpu",           0x0800, 0x0800, 0x6c3c6ebb )

ROM_REGION( 0x2000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "8.cpu",           0x0000, 0x0800, 0x5596b345 )
ROM_LOAD( "10.cpu",          0x0800, 0x0800, 0x50c7477d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",    0x0000, 0x0020, 0x63dd53a )
ROM_LOAD( "pacplus.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mspacad )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xc3dca510 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x57cb31e3 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END


ROM_START( msberzk )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x795b1d20 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x9fc2a679 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x6749a32d )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x92692429 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x7c32ed2e )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msdstorm )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xd32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xb2940b89 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x5c65865f )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msindy )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x8e234b3a )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xc7183fd2 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xa355e061 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mspacrip )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x5fb1ee61 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5aa9eb83 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xceb3785d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msyakman )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x827892bb )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x1108ba96 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msvectr )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x56560aef )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x57cb31e3 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mselton )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x30d0b19f )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fasthear )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xa8d6c227 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x28eb876c )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x50fc9966 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( heartbn2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x28eb876c )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x50fc9966 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( fastmspa )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xa8d6c227 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( chtmspa )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( mspacat2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "mspacatk.1",   0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "mspacatk.3",   0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "mspacatk.4",   0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0x5a72a7a6 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0xbd0c3e94 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( vectratk )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "mspacatk.1",   0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "mspacatk.3",   0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "mspacatk.4",   0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x56560aef )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x57cb31e3 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msdroid )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x984b3e88 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x97f5dd55 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xc0e47314 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msgrkatk )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x7143729f )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "mspacatk.6",   0x9000, 0x1000, 0x3b5db308 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xb6526896 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x043b1c9b )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( msgrkb )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x7143729f )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xb6526896 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x043b1c9b )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( abortman )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "namcopac.6e",  0x0000, 0x1000, 0xfee263b3 )
ROM_LOAD( "namcopac.6f",  0x1000, 0x1000, 0x39d1fc83 )
ROM_LOAD( "namcopac.6h",  0x2000, 0x1000, 0x02083b03 )
ROM_LOAD( "namcopac.6j",  0x3000, 0x1000, 0x7a36fe55 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x1c4ef687 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x38a22ac4 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

/*BlueJustice*/
ROM_START( bjhangp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x022c35da )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x7a444bda )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x063dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( bjhang2k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0xcba39624 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xa2b31528 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( bjhangb )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x9a8cd8a1 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xf0bd55bd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( bjhangb5 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x5ebb61c5 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xd9a055aa )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( bjhangb7 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x7961ed4e )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x417080fd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x063dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( bjhangbp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x16d9f7ce )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xceb721be )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( bjhanggp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x5807a438 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0x1fcc0cf2 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x063dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( bjhang1k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "hangly.6e",    0x0000, 0x1000, 0x5fe8610a )
ROM_LOAD( "hangly.6f",    0x1000, 0x1000, 0x73726586 )
ROM_LOAD( "hangly.6h",    0x2000, 0x1000, 0x4e7ef99f )
ROM_LOAD( "hangly.6j",    0x3000, 0x1000, 0x7f4147e6 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",    0x0000, 0x1000, 0x16d9f7ce )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",    0x0000, 0x1000, 0xfd06d923 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspac1k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x22a10392 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xadde7864 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x5cb008a9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspac2k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xa72e49df )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x2bcce741 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x4b40032a )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspac3k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x46790576 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x1708299b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xcf4798e3 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspac5k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xa72e49df )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xfd1bd093 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x07d39d56 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspac5kp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xd7ff3b21 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x4fd847aa )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x50d5fa64 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspaccm )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xf2b17834 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x43936999 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0xa15b524e )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x2bcce741 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x2a9d6f77 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspacdx )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xaf8c63ad )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x605becd7 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x435de4cb )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspacex )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xb13d0521 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xb656c0c5 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xc5d7017b )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspacexp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x728f7465 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x1a2fa3e2 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x5a400cbc )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( mspacsm )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x513f4d5c )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xef9cce5a )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0xb8f8edda )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x5d7888bc )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x14fd0ad7 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x67fb6254 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x063dd53a )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacman4k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0xadb41be2 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x74f6f4d3 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacman5k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x5865ce98 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x81cf02e1 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x063dd53a )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacman6k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x89945690 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x9268eb5a )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacman7k )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0xc181ad5d )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0xbd31be36 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacman7p )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0xc181ad5d )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x3d5861a5 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanb2 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",   0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",   0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",   0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",   0x3000, 0x1000, 0xf7d647d8 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",   0x0000, 0x1000, 0xff6308bb )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",   0x0000, 0x1000, 0xc8b456b5 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacman.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacman.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanb3 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",   0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",   0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",   0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",   0x3000, 0x1000, 0x08f7a148 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",   0x0000, 0x1000, 0x5ebb61c5 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",   0x0000, 0x1000, 0xd9a055aa )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacman.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacman.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanb4 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x5ebb61c5 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0xd9a055aa )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanb6 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",   0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",   0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",   0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",   0x3000, 0x1000, 0xbdb3d11a )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",   0x0000, 0x1000, 0x7961ed4e )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",   0x0000, 0x1000, 0x417080fd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacman.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacman.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanb8 )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x7961ed4e )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x417080fd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x063dd53a )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0xe271a166 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanbp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x3e5a31c2 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x3c099bbd )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmandx )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0xcba39624 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0xab172fc9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanex )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x3ee34b66 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0xedf92a4c )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanfp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x5ebb61c5 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0xd9a055aa )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmanfv )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0xf9b955e1 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x8ddf8eb6 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( pacmbp )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacplus.6e",   0x0000, 0x1000, 0xd611ef68 )
ROM_LOAD( "pacplus.6f",   0x1000, 0x1000, 0xc7207556 )
ROM_LOAD( "pacplus.6h",   0x2000, 0x1000, 0xae379430 )
ROM_LOAD( "pacplus.6j",   0x3000, 0x1000, 0x5a6dff7b )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5e",   0x0000, 0x1000, 0x16d9f7ce )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacplus.5f",   0x0000, 0x1000, 0x2da4dec7 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacplus.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacplus.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( petshop )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x75230eef )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0xac1c180b )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xce314b46 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

ROM_START( heartbrn )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0x0d32de5e )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x8c3e6de6 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x368cb165 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5431d4c4 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0xceb50654 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacatk )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xd16b31b7 )
ROM_LOAD( "mspacatk.2",   0x1000, 0x1000, 0x0af09d31 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0x1821ee0b )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "mspacatk.5",   0x8000, 0x1000, 0xe6e06954 )
ROM_LOAD( "u6pacatk",     0x9000, 0x1000, 0xa35b3788 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( bucaneer )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "6e.cpu",      0x0000, 0x1000, 0xfee263b3 )
ROM_LOAD( "6f.cpu",    0x1000, 0x1000, 0x39d1fc83 )
ROM_LOAD( "6h.cpu",      0x2000, 0x1000, 0x197443f8 )
ROM_LOAD( "6k.cpu",      0x3000, 0x1000, 0xc4d9169a )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "5e.cpu",      0x0000, 0x0800, 0x4060c077 )
ROM_LOAD( "5h.cpu",      0x0800, 0x0800, 0xe3861283 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "5f.cpu",      0x0000, 0x0800, 0x09f66dec )
ROM_LOAD( "5k.cpu",      0x0800, 0x0800, 0x653314e7 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

ROM_START( crash )
ROM_REGION( 0x10000, REGION_CPU1, 0 )
ROM_LOAD( "boot1",        0x0000, 0x1000, 0x04353b41 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xe03205c0 )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0xb0fa8e46 )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0xbfa4d2fe )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x12f2f224 )
ROM_LOAD( "boot6",        0x9000, 0x1000, 0x817d94e3 )

ROM_REGION( 0x1000, REGION_GFX1 , ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x0a25969b )

ROM_REGION( 0x1000, REGION_GFX2 , ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x447ea79c )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x2bc5d339 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END

ROM_START( pacmar )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	
ROM_LOAD( "pacman.6e",   0x0000, 0x1000, 0xc1e6ab10 )
ROM_LOAD( "pacman.6f",   0x1000, 0x1000, 0x1a6fb2d4 )
ROM_LOAD( "pacman.6h",   0x2000, 0x1000, 0xbcdd1beb )
ROM_LOAD( "pacman.6j",   0x3000, 0x1000, 0xb208c7dc )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5e",   0x0000, 0x1000, 0x7ccdfa59 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "pacman.5f",   0x0000, 0x1000, 0x51af9fc5 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "pacman.7f",   0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "pacman.4a",   0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

/* End PacMAME */


/*************************************
 *
 *	Driver initialization
 *
 *************************************/

static void maketrax_rom_decode(void)
{
	unsigned char *rom = memory_region(REGION_CPU1);
	int diff = memory_region_length(REGION_CPU1) / 2;


	/* patch protection using a copy of the opcodes so ROM checksum */
	/* tests will not fail */
	memory_set_opcode_base(0,rom+diff);

	memcpy(rom+diff,rom,diff);

	rom[0x0415 + diff] = 0xc9;
	rom[0x1978 + diff] = 0x18;
	rom[0x238e + diff] = 0xc9;
	rom[0x3ae5 + diff] = 0xe6;
	rom[0x3ae7 + diff] = 0x00;
	rom[0x3ae8 + diff] = 0xc9;
	rom[0x3aed + diff] = 0x86;
	rom[0x3aee + diff] = 0xc0;
	rom[0x3aef + diff] = 0xb0;
}

static DRIVER_INIT( maketrax )
{
	/* set up protection handlers */
	install_mem_read_handler(0, 0x5080, 0x50bf, maketrax_special_port2_r);
	install_mem_read_handler(0, 0x50c0, 0x50ff, maketrax_special_port3_r);

	maketrax_rom_decode();
}

static DRIVER_INIT( ponpoko )
{
	int i, j;
	unsigned char *RAM, temp;

	/* The gfx data is swapped wrt the other Pac-Man hardware games. */
	/* Here we revert it to the usual format. */

	/* Characters */
	RAM = memory_region(REGION_GFX1);
	for (i = 0;i < memory_region_length(REGION_GFX1);i += 0x10)
	{
		for (j = 0; j < 8; j++)
		{
			temp          = RAM[i+j+0x08];
			RAM[i+j+0x08] = RAM[i+j+0x00];
			RAM[i+j+0x00] = temp;
		}
	}

	/* Sprites */
	RAM = memory_region(REGION_GFX2);
	for (i = 0;i < memory_region_length(REGION_GFX2);i += 0x20)
	{
		for (j = 0; j < 8; j++)
		{
			temp          = RAM[i+j+0x18];
			RAM[i+j+0x18] = RAM[i+j+0x10];
			RAM[i+j+0x10] = RAM[i+j+0x08];
			RAM[i+j+0x08] = RAM[i+j+0x00];
			RAM[i+j+0x00] = temp;
		}
	}
}

static void eyes_decode(unsigned char *data)
{
	int j;
	unsigned char swapbuffer[8];

	for (j = 0; j < 8; j++)
	{
		swapbuffer[j] = data[(j >> 2) + (j & 2) + ((j & 1) << 2)];
	}

	for (j = 0; j < 8; j++)
	{
		char ch = swapbuffer[j];

		data[j] = (ch & 0x80) | ((ch & 0x10) << 2) |
					 (ch & 0x20) | ((ch & 0x40) >> 2) | (ch & 0x0f);
	}
}

static DRIVER_INIT( eyes )
{
	int i;
	unsigned char *RAM;

	/* CPU ROMs */

	/* Data lines D3 and D5 swapped */
	RAM = memory_region(REGION_CPU1);
	for (i = 0; i < 0x4000; i++)
	{
		RAM[i] =  (RAM[i] & 0xc0) | ((RAM[i] & 0x08) << 2) |
				  (RAM[i] & 0x10) | ((RAM[i] & 0x20) >> 2) | (RAM[i] & 0x07);
	}


	/* Graphics ROMs */

	/* Data lines D4 and D6 and address lines A0 and A2 are swapped */
	RAM = memory_region(REGION_GFX1);
	for (i = 0;i < memory_region_length(REGION_GFX1);i += 8)
		eyes_decode(&RAM[i]);
	RAM = memory_region(REGION_GFX2);
	for (i = 0;i < memory_region_length(REGION_GFX2);i += 8)
		eyes_decode(&RAM[i]);
}


static DRIVER_INIT( pacplus )
{
	pacplus_decode();
}
/*
static DRIVER_INIT( jumpshot )
{
	jumpshot_decode();
}

static DRIVER_INIT( shootbul )
{
	shootbul_decode();
}
*/

/*************************************
 *
 *	Game drivers
 *
 *************************************/

/*          rom       parent    machine   inp       init */
GAME( 1980, puckman,  0,        pacman,   pacman,   0,        ROT90,  "Namco", "PuckMan (Japan set 1)" )
GAME( 1980, puckmana, puckman,  pacman,   pacman,   0,        ROT90,  "Namco", "PuckMan (Japan set 2)" )
GAME( 1980, pacman,   puckman,  pacman,   pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man (Midway)" )
GAME( 1981, puckmod,  puckman,  pacman,   pacman,   0,        ROT90,  "Namco", "PuckMan (harder?)" )
GAME( 1981, pacmod,   puckman,  pacman,   pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man (Midway, harder)" )
GAME( 1981, hangly,   puckman,  pacman,   pacman,   0,        ROT90,  "hack", "Hangly-Man (set 1)" )
GAME( 1981, hangly2,  puckman,  pacman,   pacman,   0,        ROT90,  "hack", "Hangly-Man (set 2)" )
GAME( 1980, newpuckx, puckman,  pacman,   pacman,   0,        ROT90,  "hack", "New Puck-X" )
GAME( 1981, pacheart, puckman,  pacman,   pacman,   0,        ROT90,  "hack", "Pac-Man (Hearts)" )
GAME( 1981, piranha,  puckman,  pacman,   mspacman, 0,        ROT90,  "hack", "Piranha" )
GAME( 1982, pacplus,  0,        pacman,   pacman,   pacplus,  ROT90,  "[Namco] (Midway license)", "Pac-Man Plus" )
GAME( 1981, mspacman, 0,        mspacman, mspacman, 0,        ROT90,  "Midway", "Ms. Pac-Man" )
GAME( 1981, mspacmab, mspacman, pacman,   mspacman, 0,        ROT90,  "bootleg", "Ms. Pac-Man (bootleg)" )
GAME( 1981, mspacmat, mspacman, mspacman, mspacman, 0,        ROT90,  "hack", "Ms. Pac Attack" )
GAME( 1981, mspacpls, mspacman, pacman,   mspacman, 0,        ROT90,  "hack", "Ms. Pac-Man Plus" )
GAME( 1981, pacgal,   mspacman, pacman,   mspacman, 0,        ROT90,  "hack", "Pac-Gal" )
GAME( 1995, mschamp,  mspacman, mschamp,  mschamp,  0,        ROT90,  "hack", "Ms. Pacman Champion Edition / Super Zola Pac Gal" )

GAME( 1981, crush,    0,        pacman,   maketrax, maketrax, ROT90,  "Kural Samno Electric", "Crush Roller (Kural Samno)" )
GAME( 1981, crush2,   crush,    pacman,   maketrax, 0,        ROT90,  "Kural Esco Electric", "Crush Roller (Kural Esco - bootleg?)" )
GAME( 1981, crush3,   crush,    pacman,   maketrax, eyes,     ROT90,  "Kural Electric", "Crush Roller (Kural - bootleg?)" )
GAME( 1981, maketrax, crush,    pacman,   maketrax, maketrax, ROT270, "[Kural] (Williams license)", "Make Trax" )
GAME( 1981, mbrush,   crush,    pacman,   mbrush,   0,        ROT90,  "bootleg", "Magic Brush" )
GAME( 1981, paintrlr, crush,    pacman,   paintrlr, 0,        ROT90,  "bootleg", "Paint Roller" )
GAME( 1982, ponpoko,  0,        pacman,   ponpoko,  ponpoko,  ROT0,   "Sigma Ent. Inc.", "Ponpoko" )
GAME( 1982, ponpokov, ponpoko,  pacman,   ponpoko,  ponpoko,  ROT0,   "Sigma Ent. Inc. (Venture Line license)", "Ponpoko (Venture Line)" )
GAME( 1982, eyes,     0,        pacman,   eyes,     eyes,     ROT90,  "Digitrex Techstar (Rock-ola license)", "Eyes (Digitrex Techstar)" )
GAME( 1982, eyes2,    eyes,     pacman,   eyes,     eyes,     ROT90,  "Techstar Inc. (Rock-ola license)", "Eyes (Techstar Inc.)" )
GAMEX(1983, mrtnt,    0,        pacman,   mrtnt,    eyes,     ROT90,  "Telko", "Mr. TNT", GAME_WRONG_COLORS )
GAME( 1985, lizwiz,   0,        pacman,   lizwiz,   0,        ROT90,  "Techstar (Sunn license)", "Lizard Wizard" )
/*
GAME( 1983, theglobp, suprglob, theglobp, theglobp, 0,        ROT90,  "Epos Corporation", "The Glob (Pac-Man hardware)" )
GAME( 1984, beastf,   suprglob, theglobp, theglobp, 0,        ROT90,  "Epos Corporation", "Beastie Feastie" )
*/
GAME( 1982, dremshpr, 0,        dremshpr, dremshpr, 0,        ROT270, "Sanritsu", "Dream Shopper" )
GAME( 1983, vanvan,   0,        vanvan,   vanvan,   0,        ROT270, "Sanritsu", "Van-Van Car" )
GAME( 1983, vanvank,  vanvan,   vanvan,   vanvans,  0,        ROT270, "Karateco", "Van-Van Car (Karateco)" )
GAMEX(1982, alibaba,  0,        alibaba,  alibaba,  0,        ROT90,  "Sega", "Ali Baba and 40 Thieves", GAME_WRONG_COLORS | GAME_UNEMULATED_PROTECTION )
//GAME( 1985, jumpshot, 0,        pacman,   jumpshot, jumpshot, ROT90,  "Bally Midway", "Jump Shot" )
//GAME( 1985, shootbul, 0,        pacman,   shootbul, shootbul, ROT90,  "Bally Midway", "Shoot the Bull" )

/* PacMAME Drivers */
/* I have added the parent of each set to each game, this is needed for my PacTray frontend */
/* and the automatic update feature to work. Once a new PacMAME is released, PacTray will issue */
/* the 'pacmame -listclones > list.dat' and rebuild the pacmame.lst file from this. The pacmame.lst */
/* file contains the 'name,rom name, screenshot, cabinet, flyer, hi score format' list. */
/* The name & rom name are pretty straight forward, the others are based off of the parent set */
/* name, for example, pacbell is a clone of pacman, therefore the parent is pacman, so, when */
/* PacTray rebuilds the pacmame.lst file the entry for pacbell will look like this */

/* Pacman (The Bell),pacbell,pacbell.png,pacman.png,pacman.png,pacman */
/*                                           ^-----------^--------^------- from parent name */
/* Make sense ? I hope. */
/* I know other emus remove this (ie EmuPlus, which kicks ass BTW), but if there is another way */
/* to accomplish what I need, then I can change it */
/* I've updated FRONTHLP.C so that the -listfull display will include the 'clone of' display */

/*    Year  Name      Parent    Machine  Input     Init	     Monitor Company    Full Name */
GAME( 1981, hanglyb,  puckman,   pacman,  pacman,   0,        ROT90,  "hack", "Hangly-Man (Hearts)" )

GAME( 2000, baby2,    puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Baby Pacman 2 (Alt)" )
GAME( 2000, baby3,    puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Baby Pacman 3 (Alt)" )
GAME( 2000, baby4,    puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Pacman (Baby Maze 4)" )
GAME( 2000, brakman,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Brakman" )
GAME( 2000, caterpil, puckman,   pacman,  pacman,   0,        ROT90,  "Phi", "Caterpillar" )
GAME( 2000, chthang,  puckman,   fastpac, pacman,   0,        ROT90,  "[Namco] (Midway license)", "Hangly Man (Cheat)" )
GAME( 2000, chtmsatk, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Plus (Cheat)" )
GAME( 2000, chtmspa,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man (Cheat)" )
GAME( 2000, chtpac,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man (Cheat)" )
GAME( 2000, chtpman2, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Puck-Man II (Cheat)" )
GAME( 2000, chtpop,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man / Popeye (Cheat)" )
GAME( 2000, chtpuck,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Puck-Man (Cheat)" )
GAME( 2001, cookiem,  mspacman, pacman,  mspacman, 0,        ROT90,  "Nic", "Cookie-Mon!" )
GAME( 2002, crash,    mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Crash" )
GAME( 2000, crazypac, puckman,   pacman,  pacman,   0,        ROT90,  "Tim Appleton a.k.a. medragon", "Crazy Pac" )
GAME( 2000, dizzy,    puckman,   pacman,  pacman,   0,        ROT90,  "Tim Appleton a.k.a. medragon", "Dizzy Ghost - A Reversal of Roles" )
GAME( 2000, eltonpac, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Elton Pac" )
GAME( 2000, europac,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Euro Pac" )
GAME( 2000, fasthang, puckman,   fastpac, pacman,   0,        ROT90,  "[Namco] (Midway license)", "Hangly Man (Fast)" )
GAME( 2000, fasthear, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Heart Burn (Fast)" )
GAME( 2000, fastmspa, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man (Fast)" )
GAME( 2000, fastpac,  puckman,   fastpac, pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man (Fast)" )
GAME( 2000, fastplus, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Plus (Fast)" )
GAME( 2000, fastpop,  puckman,   fastpac, pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man / Popeye (Fast)" )
GAME( 2000, fastpuck, puckman,   fastpac, pacman,   0,        ROT90,  "[Namco] (Midway license)", "Puck-Man (Fast)" )
GAME( 2000, faststrm, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Desert Storm (Fast)" )
GAME( 2000, fpnleash, mspacman, pacman,  mspacman, 0,        ROT90,  "Peter Storey", "Pacman Unleashed (Flat)" )
GAME( 2000, fstmsatk, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Attack (Fast)" )
GAME( 2000, fstpman2, puckman,   fastpac, pacman,   0,        ROT90,  "[Namco] (Midway license)", "Puck-Man II (Fast)" )
GAME( 2000, hanglyad, puckman,   pacman,  pacman,   0,        ROT90,  "hack", "Hangly-Man (set 1) After Dark" )
GAME( 2000, heartbn2, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Heart Burn" )
GAME( 2000, heartbrn, mspacman, pacman,  mspacman, 0,        ROT90,  "TwoBit Score", "Ms. Heart Burn (TwoBit Score)" )
GAME( 2000, hearts,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Hearts Bootleg)" )
GAME( 2000, mazeman,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Maze Man" )
GAME( 2000, mrpacman, mspacman, pacman,  mspacman, 0,        ROT90,  "Tim Appleton a.k.a. medragon", "Mr. Pac-Man - Another Kind of Role Reversal" )
GAME( 2000, ms2600,   mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman 2600" )
GAME( 2000, msatk2ad, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Plus / Attack After Dark" )
GAME( 2000, msatkad,  mspacman, pacman,  mspacman, 0,        ROT90,  "hack", "Ms. Pac-Man Plus After Dark" )
GAME( 2000, msbaby,   mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Baby Pacman" )
GAME( 2000, msbaby1,  mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Baby Pacman (Alt)" )
GAME( 2000, msberzk,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Berzerk" )
GAME( 2000, msdstorm, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Desert Storm" )
GAME( 2000, mselton,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Elton" )
GAME( 2000, msf1pac,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. F1 Pac-Man" )
GAME( 2000, mshangly, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Hangly Man" )
GAME( 2000, mshearts, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Hearts" )
GAME( 2000, msindy,   mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Indy" )
GAME( 2000, msmini,   mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pac-Mini" )
GAME( 2000, msminia,  mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pac-Attack Mini" )
GAME( 2000, msnes4a,  mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman SNES (Tall Alternate)" )
GAME( 2000, msnes62,  mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman NES (Set 2)" )
GAME( 2000, msnes63,  mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman NES (Set 3)" )
GAME( 2000, msnes6m,  mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman NES" )
GAME( 2000, msnes6m2, mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman SNES (New Mazes)" )
GAME( 2000, msnes6m3, mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman SNES (Regular / Tall" )
GAME( 2000, msnes6m4, mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman SNES (Regular)" )
GAME( 2000, mspac6m,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man 6M (Six Maze)" )
GAME( 2000, mspacad,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man After Dark" )
GAME( 2000, mspacat2, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Plus / Attack" )
GAME( 2000, mspacatb, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man Plus (Set B)" )
GAME( 2000, mspacren, mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman Renaissance" )
GAME( 2000, mspacrip, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Mortem" )
GAME( 2000, mspc6mad, mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Ms. Pacman After Dark (6 Mazes)" )
GAME( 2000, msrumble, mspacman, pacman,  mspacman, 0,        ROT90,  "Tim Appleton a.k.a. medragon", "Ms. Pac Rumble" )
GAME( 2000, mssilad,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pac-Man After Dark (Sil)" )
GAME( 2000, msultra,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacFan", "Ms. Ultra Pacman" )
GAME( 2000, msvctr6m, mspacman, pacman,  mspacman, 0,        ROT90,  "T-Bone", "Vector Ms. Pacman (6 Mazes)" )
GAME( 2000, msvectr,  mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Vector Ms. Pac-Man" )
GAME( 2000, msyakman, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Yak Man" )
GAME( 2000, pac2600,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac 2600" )
GAME( 2000, pacbaby,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Baby Maze 1)" )
GAME( 2000, pacbaby2, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Baby Maze 2)" )
GAME( 2000, pacbaby3, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Baby Maze 3)" )
GAME( 2000, pacbell,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Bell)" )
GAME( 2000, pacelec,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Electric Cowboy)" )
GAME( 2000, pacinvis, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Original Inviso)" )
GAME( 2000, pacjail,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Jail)" )
GAME( 2000, pacjr1,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman Jr. (Maze 1)" )
GAME( 2000, pacjr2,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman Jr. (Maze 2)" )
GAME( 2000, pacjr3,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman Jr. (Maze 3)" )
GAME( 2000, pacjr4,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman Jr. (Maze 4)" )
GAME( 2000, pacm255,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man (Midway 255th Maze)" )
GAME( 2000, pacman3d, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man 3D" )
GAME( 2000, pacman6,  mspacman, pacman,  mspacman, 0,        ROT90,  "Sil", "Pacman 6" )
GAME( 2000, pacmini,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Mini Pac-Man" )
GAME( 2000, pacmini2, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Mini Pac-Man 2" )
GAME( 2000, pacmn6m2, puckman,   pacman,  pacman,   0,        ROT90,  "Sil", "Pacman 2000 (Set 2)" )
GAME( 2000, pacms1,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Maze 1)" )
GAME( 2000, pacms2,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Maze 2)" )
GAME( 2000, pacms3,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Maze 3)" )
GAME( 2000, pacms4,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Maze 4)" )
GAME( 2000, pacmsa1,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Attack 1)" )
GAME( 2000, pacmsa2,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Attack 2)" )
GAME( 2000, pacmsa3,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Attack 3)" )
GAME( 2000, pacmsa4,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Ms. Pacman Attack 4)" )
/*
GAME( 2000, pacmulti, pacman,   pacman,  pacman,   0,        ROT90,  "Sil", "PacMulti (Pacman)" )
*/
GAME( 2000, pacn255,  0,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man (Namco 255th Maze)" )
GAME( 2000, pacplusc, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man Plus (Unencrypted)" )
GAME( 2000, pacshuf,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Shuffle)" )
GAME( 2000, pacspeed, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Speedy)" )
GAME( 2000, pacstrm,  puckman,   pacman,  pacman,   0,        ROT90,  "PacMAME", "Desert Storm Pac-Man" )
GAME( 2000, pacweird, puckman,   pacman,  pacman,   0,        ROT90,  "Staizitto", "Pacman (Six Map Weird)" )
GAME( 2000, pcrunchy, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pacman (Crunchy)" )
GAME( 2000, pengman,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pengo Man" )
GAME( 2000, plusfast, puckman,   plusfast,pacman,   pacplus,  ROT90,  "[Namco] (Midway license)", "Pac-Man Plus (Fast)" )
GAME( 2000, pmad,     puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man After Dark" )
GAME( 2000, pmad00,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man After Dark 2000" )
GAME( 2000, pmad6m,   puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Pacman 2000 After Dark" )
GAME( 2000, pmada,    puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man After Dark (Alternate)" )
GAME( 2000, pmanalt,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man Vertical Tunnels (Cheat)" )
GAME( 2000, pmanaltm, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Pac-Man (Midway Vertical Tunnels)" )
GAME( 2000, pplusad,  puckman,   pacman,  pacman,   pacplus,  ROT90,  "[Namco] (Midway license)", "Pac-Man Plus After Dark" )
GAME( 2000, puckman2, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Puck-Man II" )
GAME( 2000, puckren,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Puckman Renaissance" )
GAME( 2000, puckrenc, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Puckman Renaissance (Cheat)" )
GAME( 2000, punleash, mspacman, pacman,  mspacman, 0,        ROT90,  "Peter Storey", "Pacman Unleashed" )
GAME( 2000, roboman,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Robo Man" )
GAME( 2000, snakeyes, puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Pacman 2000 (Snake Eyes Mazes)")
GAME( 2000, snowpac,  puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Snowy Day Pacman" )
GAME( 2000, sueworld, mspacman, pacman,  mspacman, 0,        ROT90,  "PacFay", "Sue's World" )
GAME( 2000, sumelton, puckman,   pacman,  pacman,   0,        ROT90,  "Staizitto", "Summertime Elton" )
GAME( 2000, tbone,    puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Pacman 2000 (T-Bone Mazes)")
GAME( 2000, ultra2,   puckman,   pacman,  pacman,   0,        ROT90,  "PacMAME", "Ultra Pacman (TwoBit Score Mazes)" )
GAME( 2000, ultrapac, puckman,   pacman,  pacman,   0,        ROT90,  "PacFan", "UltraPac" )
GAME( 2000, vcrunchy, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Crunchy)" )
GAME( 2000, vecbaby,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Baby Maze 1)" )
GAME( 2000, vecbaby2, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Baby Maze 2)" )
GAME( 2000, vecbaby3, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Baby Maze 3)" )
GAME( 2000, vectplus, puckman,   pacman,  pacman,   pacplus,  ROT90,  "[Namco] (Midway license)", "Vector Pac-Man Plus" )
GAME( 2000, vectr6m,  puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Vector Pacman 2000")
GAME( 2000, vectr6tb, puckman,   pacman,  pacman,   0,        ROT90,  "T-Bone", "Vector Pacman 2000 (T-Bone Mazes)")
GAME( 2000, vectratk, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Vector Attack" )
GAME( 2000, vectrpac, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pac-Man" )
GAME( 2000, vectxens, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Xens Revenge" )
GAME( 2000, vhangly,  puckman,   pacman,  pacman,   0,        ROT90,  "hack", "Vector Hangly-Man (set 1)" )
GAME( 2000, vpacbell, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Bell)" )
GAME( 2000, vpacelec, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Electric Cowboy)" )
GAME( 2000, vpacjail, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Jail)" )
GAME( 2000, vpacms1,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Maze 1)" )
GAME( 2000, vpacms2,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Maze 2)" )
GAME( 2000, vpacms3,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Maze 3)" )
GAME( 2000, vpacms4,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Maze 4)" )
GAME( 2000, vpacmsa1, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Attack 1)" )
GAME( 2000, vpacmsa2, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Attack 2)" )
GAME( 2000, vpacmsa3, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Attack 3)" )
GAME( 2000, vpacmsa4, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Ms. Pacman Attack 4)" )
GAME( 2000, vpacshuf, puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Shuffle)" )
GAME( 2000, vpspeed,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Vector Pacman (Speedy)" )
GAME( 2000, xensad,   puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Xens Revenge After Dark" )
GAME( 2000, xensrev,  puckman,   pacman,  pacman,   0,        ROT90,  "[Namco] (Midway license)", "Xens Revenge" )
GAME( 2000, zolaad,   mspacman, zolapac, zolapac,  0,        ROT90,  "PacMAME", "Super Zola Pac Gal After Dark" )
GAME( 2000, zolaatk,  mspacman, zolapac, zolapac,  0,        ROT90,  "PacMAME", "Super Zola Pac Gal Attack" )
GAME( 2000, zolamaze, mspacman, pacman,  mspacman, 0,        ROT90,  "PacMAME", "Ms. Pacman (Zola Mazes)" )
GAME( 2000, zolapac,  mspacman, zolapac, zolapac,  0,        ROT90,  "PacMAME", "Super Zola Pac Gal" )
GAME( 2000, zolaslow, mspacman, zolapac, zolapac,  0,        ROT90,  "PacMAME", "Super Zola Pac Gal (Slow)" )
GAME( 2000, zolatunl, mspacman, zolapac, zolapac,  0,        ROT90,  "PacMAME", "Super Zola Pac Gal (Tunnels)" )
GAME( 2000, zolavect, mspacman, zolapac, zolapac,  0,        ROT90,  "PacMAME", "Vector Super Zola Pac Gal" )
/*
GAME( 1995, mschamp,  mspacman,  mschamp, mschamp,  mschamp,  ROT90,  "PacMAME", "Ms. Pacman Championship Edition" )
*/
GAME( 2001, msdroid,  mspacman, pacman,  mspacman, 0,        ROT90,  "Grendal74", "Ms. Pacman Android" )
GAME( 2001, msgrkb,   mspacman, pacman,  mspacman, 0,        ROT90,  "Grendal74", "Ms. Pacman Greek" )
GAME( 2001, msgrkatk, mspacman, pacman,  mspacman, 0,        ROT90,  "Grendal74", "Ms. Pacman Greek Bootleg" )

GAME( 2001, abortman, 0,   pacman,  pacman,   0,        ROT90,  "Paul Copeland", "Abortman" )

/*BlueJustice*/
GAME( 2001, bjhangp,   puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man Plus" )
GAME( 2001, bjhang2k,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man 2000" )
GAME( 2001, bjhangb,   puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man Babies" )
GAME( 2001, bjhangb5,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man Babies 5000" )
GAME( 2001, bjhangb7,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man Babies 7000" )
GAME( 2001, bjhangbp,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man Blueplus" )
GAME( 2001, bjhanggp,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man Greenplus" )
GAME( 2001, bjhang1k,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Hangly-Man 1000" )
GAME( 2001, mspac1k,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman 1000" )
GAME( 2001, mspac2k,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman 2000" )
GAME( 2001, mspac3k,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman 3000" )
GAME( 2001, mspac5k,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman 5000" )
GAME( 2001, mspac5kp,  mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman 5000 Plus" )
GAME( 2001, mspaccm,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman Crazy Mazes" )
GAME( 2001, mspacdx,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman Deluxe" )
GAME( 2001, mspacex,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman Extra" )
GAME( 2001, mspacexp,  mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman Extra Plus" )
GAME( 2001, mspacsm,   mspacman, pacman,   mspacman, 0,        ROT90,  "BlueJustice", "Ms. Pacman Strange Mazes" )
GAME( 2001, pacman4k,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Plus 4000" )
GAME( 2001, pacman5k,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Plus 5000" )
GAME( 2001, pacman6k,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Plus 6000" )
GAME( 2001, pacman7k,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Plus 7000" )
GAME( 2001, pacman7p,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man 7000 Plus" )
GAME( 2001, pacmanb2,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Pac-Man Babies 2000" )
GAME( 2001, pacmanb3,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Pac-Man Babies 3000" )
GAME( 2001, pacmanb4,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Plus Babies 4000" )
GAME( 2001, pacmanb6,  puckman,  pacman,   pacman,   0,        ROT90,  "BlueJustice", "Pac-Man Babies 6000" )
GAME( 2001, pacmanb8,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Plus Babies 8000" )
GAME( 2001, pacmanbp,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Blueplus" )
GAME( 2001, pacmandx,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Plus Deluxe" )
GAME( 2001, pacmanex,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Extra Plus" )
GAME( 2001, pacmanfp,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Fever Plus" )
GAME( 2001, pacmanfv,  0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Fever" )
GAME( 2001, pacmbp,    0,  pacman,   pacman,   pacplus,  ROT90,  "BlueJustice", "Pac-Man Babies Plus" )

GAME (2001, petshop,  mspacman, pacman,   mspacman, 0,        ROT90,  "Hack", "Pet Shop Freak-Out!")
GAME (2000, pacatk,   mspacman, pacman,   mspacman, 0,        ROT90,  "David Widel", "Ms. Pac-Man Plus (Fruit)" )
GAME (1981, bucaneer, puckman,  pacman,   mspacman, 0,        ROT90,  "PacMAME", "Bucaneer" )
GAME( 2001, pacmar,   puckman,  pacman,   pacman,   0,        ROT90,  "PacMAME", "Mario Pacman" )

/*********************************************************************
SUPER ABC & MULTIPAC
*********************************************************************/

//   Multipac 1.4 & 1.5 Driver
//   I've made the driver so that It resets the hardware as soon as the banks
//   change. This means that the games all _work_, wether that's what the
//   original does? beats me!

static INTERRUPT_GEN( multipac_interrupt )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	/* speed up cheat */
	if (speedcheat)
	{
		if (readinputport(4) & 1)	/* check status of the fake dip switch */
		{
			/* activate the cheat */
			RAM[0x180b] = 0x01;
			RAM[0x1ffd] = 0xbd;
		}
		else
		{
			/* remove the cheat */
			RAM[0x180b] = 0xbe;
			RAM[0x1ffd] = 0x00;
		}
	}

	if (readinputport(5) & 1) {
		memset(dirtybuffer,1,videoram_size);
		RAM[0xa000]=0;
		RAM[0xa001]=0;
		multipac_bankaddress = 0x10000;
		memset (dirtybuffer,1,videoram_size);
		cpu_setbank(1, &RAM[multipac_bankaddress]);
//TODO
//		cpu_set_reg(1, 0x0000); // Force 0 into PC
		gfx_bank_zero();        // Force GFX Bank 0 (Menu Bank)      
		// z80_reset(); This Crashes MAME (boo-hoo!)
	}

	irq0_line_hold();
}

	/* Menu Button */
	// If menu button pressed, set bank to 0, and reset the cpu
	// THIS DOESN'T WORK!!!!

	// I need a way to force gfx_bank to be zero
	// and to disable interrupts, and turn off sound.
	// beyond my reach atm...

INPUT_PORTS_START( multipac )
PORT_START	/* IN0 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
PORT_DIPSETTING(    0x00, DEF_STR( On ) )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

PORT_START	/* IN1 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
PORT_SERVICE( 0x10, IP_ACTIVE_LOW ) 
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_BUTTON1 | IPF_PLAYER1 ) // Button 1 maps to start button also.
// makes lizard wizard easier to play
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
PORT_DIPNAME(0x80, 0x80, DEF_STR( Cabinet ) )
PORT_DIPSETTING(   0x80, DEF_STR( Upright ) )
PORT_DIPSETTING(   0x00, DEF_STR( Cocktail ) )

PORT_START	/* DSW 1 */
PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
PORT_DIPSETTING(    0x00, "1" )
PORT_DIPSETTING(    0x04, "2" )
PORT_DIPSETTING(    0x08, "3" )
PORT_DIPSETTING(    0x0c, "5" )
PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
PORT_DIPSETTING(    0x00, "10000" )
PORT_DIPSETTING(    0x10, "15000" )
PORT_DIPSETTING(    0x20, "20000" )
PORT_DIPSETTING(    0x30, "None" )
PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
PORT_DIPSETTING(    0x40, "Normal" )
PORT_DIPSETTING(    0x00, "Hard" )
PORT_DIPNAME( 0x80, 0x80, "Ghost Names" )
PORT_DIPSETTING(    0x80, "Normal" )
PORT_DIPSETTING(    0x00, "Alternate" )

PORT_START	/* DSW 2 */
PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )

PORT_START	/* FAKE */
/* This fake input port is used to get the status of the fire button */
/* and activate the speedup cheat if it is. */
PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Speedup Cheat", KEYCODE_LCONTROL, JOYCODE_1_BUTTON1 )
PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
PORT_DIPSETTING(    0x01, DEF_STR( On ) )

PORT_START      /* MENU BUTTON */
/* This fake input port is used to get the status of the fire button */
/* and activate the speedup cheat if it is. */
PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME, "Menu Button", KEYCODE_RCONTROL, JOYCODE_1_BUTTON1 )
PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
PORT_DIPSETTING(    0x01, DEF_STR( On ) )

INPUT_PORTS_END


static struct GfxDecodeInfo multipac_gfxdecodeinfo[] =
{
	{ REGION_GFX1, 0x00000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x01000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x02000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x03000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x04000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x05000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x06000, &tilelayout,   4*32, 32 }, //Pacman Plus//
	{ REGION_GFX1, 0x07000, &spritelayout, 4*32, 32 }, //Pacman Plus//
	{ REGION_GFX1, 0x08000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x09000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x0a000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x0b000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x0c000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x0d000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x0e000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x0f000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x10000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x11000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x12000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x13000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x14000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x15000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x16000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x17000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x18000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x19000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x1a000, &tilelayout,   8*32, 32 }, // Crush Roller //
	{ REGION_GFX1, 0x1b000, &spritelayout, 8*32, 32 }, // Crush Roller //
	{ REGION_GFX1, 0x1c000, &tilelayout,   12*32, 32 }, // Liz Wiz //
	{ REGION_GFX1, 0x1d000, &spritelayout, 12*32, 32 }, // Liz Wiz //
	{ REGION_GFX1, 0x1e000, &tilelayout,   0, 32 }, // Unused? //
	{ REGION_GFX1, 0x1f000, &spritelayout, 0, 32 }, // Unused? //                      

	{ -1 } /* end of array */
};


static void init_multipac(void)
{
	int i;
	unsigned char *RAM;

	/* Clear Bank Swap addresses */
	for (i=0; i<8; i++) {
		RAM = memory_region(REGION_CPU1);
		RAM[0xa000+i] = 0;
	}

	/* And clear Sprites */
	for (i=0; i<8; i++) {
		RAM = memory_region(REGION_CPU1);
		RAM[0x5060+i*2]=0;
		RAM[0x5061+i*2]=0;
	}

	multipac_bankaddress = 0x10000;
	cpu_setbank(1, &RAM[multipac_bankaddress]);
}

static WRITE_HANDLER( multipac_bankswitch_w)
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	if (data != 0) {
		if (data != multipac_bankaddress) {
			if (data < 0x20) // Check for valid codes...
				multipac_bankaddress = 0x10000+ ((data & 0x10) / 0x10 + (data & 0x0f) *2) * 0x4000;
			else
				multipac_bankaddress = 0x10000;
			cpu_setbank(1, &RAM[multipac_bankaddress]);

			if (RAM[0xa007]==0)
				cpu_setbank(2, &RAM[multipac_bankaddress+0x4000]);
		}

	}
}

static WRITE_HANDLER( multipac_a000_w )
{
	unsigned char *RAM = memory_region(REGION_CPU1);

	if (data != multipac_bankaddress) {
		if (data < 0x20) // Check for valid codes...
			multipac_bankaddress = 0x10000 + ((data & 0x10) / 0x10 + (data & 0x0f) *2) * 0x4000;
		else
			multipac_bankaddress = 0x10000;
		cpu_setbank(1, &RAM[multipac_bankaddress]);

		if (RAM[0xa007]==0)
			cpu_setbank(2, &RAM[multipac_bankaddress+0x4000]);
//TODO
//		cpu_set_reg(1,0x0000); // Set PC <- 0000
	} 
}


static MEMORY_READ_START( multipac_readmem )
{ 0x0000, 0x3fff, MRA_BANK1 },
{ 0x4000, 0x47ff, MRA_RAM },	/* video and color RAM */
{ 0x4c00, 0x4fff, MRA_RAM },    /* including sprite codes at 4ff0-4fff */
{ 0x5000, 0x503f, input_port_0_r },	/* IN0 */
{ 0x5040, 0x507f, input_port_1_r },	/* IN1 */
{ 0x5080, 0x50bf, input_port_2_r },	/* DSW1 */
{ 0x50c0, 0x50ff, input_port_3_r },	/* DSW2 */
//      { 0x8000, 0x9fff, MRA_ROM },    /* Ms. Pac-Man */
{ 0x8000, 0x9fff, MRA_BANK2 },  /* Ms. Pac-Man, LizWiz */
// to implement bank switching
// { 0xa000, 0xafff, MRA_RAM },    /* Multi Pac : Must it be RAM? */
{ 0xf000, 0xfffe, MRA_RAM },
MEMORY_END

static MEMORY_WRITE_START( multipac_writemem )
{ 0x0000, 0x3fff, MWA_ROM },
{ 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
{ 0x4400, 0x47ff, colorram_w, &colorram },
{ 0x4900, 0x49ff, MWA_RAM }, /* MULTIPAC */
{ 0x4c00, 0x4fef, MWA_RAM },
{ 0x4ff0, 0x4fff, MWA_RAM, &spriteram, &spriteram_size },
{ 0x5000, 0x5000, interrupt_enable_w },
{ 0x5001, 0x5001, pengo_sound_enable_w },
{ 0x5002, 0x5002, MWA_NOP },
{ 0x5003, 0x5003, pengo_flipscreen_w },
{ 0x5004, 0x5005, pacman_leds_w },
// 	{ 0x5006, 0x5006, pacman_coin_lockout_global_w },	this breaks many games
{ 0x5007, 0x5007, pacman_coin_counter_w },
{ 0x5040, 0x505f, pengo_sound_w, &pengo_soundregs },
{ 0x5060, 0x506f, MWA_RAM, &spriteram_2 },
//      { 0x50c0, 0x50c0, watchdog_reset_w },
{ 0x8000, 0x9fff, MWA_ROM },    /* Ms. Pac-Man */
{ 0xa000, 0xa000, multipac_a000_w },
{ 0xa001, 0xa001, multipac_gfxbank_w },
{ 0xa002, 0xa002, multipac_color_w },
{ 0xa003, 0xa00f, MWA_RAM },    /* Used by Multipac */
{ 0xc000, 0xc3ff, videoram_w }, /* mirror address for video ram, */
{ 0xc400, 0xc7ef, colorram_w }, /* used to display HIGH SCORE and CREDITS */
{ 0xf000, 0xfffe, MWA_RAM },
{ 0xffff, 0xffff, MWA_NOP },    /* Eyes writes to this location to simplify code */
MEMORY_END

static MACHINE_DRIVER_START( multipac )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(multipac_readmem,multipac_writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(multipac_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(pacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(multipac_gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(64)
	MDRV_COLORTABLE_LENGTH(4*128)
//TODO
//	multipac_vh_convert_color_prom,
	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END

ROM_START( multi15 )
ROM_REGION( 0x90000, REGION_CPU1, 0 )   /* 8 banks of 64k for code */
ROM_LOAD( "mpexe15.bin",  0x10000, 0x80000, 0xeb181a29 )
ROM_COPY( REGION_CPU1, 0x10000, 0x0000, 0x4000)

ROM_REGION( 0x20000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "mpgfx15.bin",  0x0000, 0x20000, 0x012fb9ec )

ROM_REGION( 0x2420, REGION_PROMS, 0)
ROM_LOAD( "pal_7f.bin",   0x1000, 0x0100, 0x40a5c3d9 )
ROM_COPY( REGION_PROMS,   0x1000, 0x0000, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1020, 0x0010, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1040, 0x0020, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1060, 0x0030, 0x0010 )

ROM_LOAD( "clut_4a.bin",  0x0040, 0x0400, 0x562a66de )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )  /* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

ROM_START( multi14 )
ROM_REGION( 0x90000, REGION_CPU1, 0 )   /* 8 banks of 64k for code */
ROM_LOAD( "mpac1_4.bin", 0x10000, 0x80000, 0x23de184d )
ROM_COPY( REGION_CPU1,   0x10000, 0x0000, 0x4000)

ROM_REGION( 0x20000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "mpgfx15.bin",  0x0000, 0x20000, 0x012fb9ec )

ROM_REGION( 0x2420, REGION_PROMS, 0)
ROM_LOAD( "pal_7f.bin",   0x1000, 0x0100, 0x40a5c3d9 )
ROM_COPY( REGION_PROMS,   0x1000, 0x0000, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1020, 0x0010, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1040, 0x0020, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1060, 0x0030, 0x0010 )

ROM_LOAD( "clut_4a.bin",  0x0040, 0x0400, 0x562a66de )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )  /* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

ROM_START( multi13 )
ROM_REGION( 0x90000, REGION_CPU1, 0 )   /* 8 banks of 64k for code */
ROM_LOAD( "mp.bin",     0x10000, 0x80000, 0x0aca3ce9 )
ROM_COPY( REGION_CPU1,   0x10000, 0x0000, 0x4000)

ROM_REGION( 0x20000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "mpgfx.bin",    0x0000, 0x20000, 0x012fb9ec )

ROM_REGION( 0x2420, REGION_PROMS, 0)
ROM_LOAD( "pal_7f.bin",   0x1000, 0x0100, 0x40a5c3d9 )
ROM_COPY( REGION_PROMS,   0x1000, 0x0000, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1020, 0x0010, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1040, 0x0020, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1060, 0x0030, 0x0010 )

ROM_LOAD( "clut_4a.bin",  0x0040, 0x0400, 0x562a66de )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )  /* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

ROM_START( multi11 )
ROM_REGION( 0x90000, REGION_CPU1, 0 )   /* 8 banks of 64k for code */
ROM_LOAD( "mpac1_1.bin", 0x10000, 0x80000, 0x0fa787fe )
ROM_COPY( REGION_CPU1,   0x10000, 0x0000, 0x4000)

ROM_REGION( 0x20000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "mpgfx.bin",   0x0000, 0x20000, 0x012fb9ec )

ROM_REGION( 0x2420, REGION_PROMS, 0)
ROM_LOAD( "pal_7f.bin",   0x1000, 0x0100, 0x40a5c3d9 )
ROM_COPY( REGION_PROMS,   0x1000, 0x0000, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1020, 0x0010, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1040, 0x0020, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1060, 0x0030, 0x0010 )

ROM_LOAD( "clut_4a.bin",  0x0040, 0x0400, 0x562a66de )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )  /* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

ROM_START( multi10 )
ROM_REGION( 0x90000, REGION_CPU1, 0 )   /* 8 banks of 64k for code */
ROM_LOAD( "master.bin",  0x10000, 0x80000, 0x9abb7d01 )
ROM_COPY( REGION_CPU1,   0x10000, 0x0000, 0x4000)

ROM_REGION( 0x20000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "master.gfx",   0x0000, 0x20000, 0x072b216f )

ROM_REGION( 0x2420, REGION_PROMS, 0)
ROM_LOAD( "pal_7f.bin",   0x1000, 0x0100, 0x40a5c3d9 )
ROM_COPY( REGION_PROMS,   0x1000, 0x0000, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1020, 0x0010, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1040, 0x0020, 0x0010 )
ROM_COPY( REGION_PROMS,   0x1060, 0x0030, 0x0010 )

ROM_LOAD( "clut_4a.bin",  0x0040, 0x0400, 0x562a66de )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )  /* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END


/*          rom       parent    machine   inp       init */
GAME( 1998, multi15,  0,        multipac, multipac, multipac,        ROT90,  "[Various] Clay Cowgill Hack", "Multipac (v1.5)" )
GAME( 1998, multi14,  multi15,  multipac, multipac, multipac,        ROT90,  "[Various] Clay Cowgill Hack", "Multipac (v1.4)" )
GAME( 1998, multi13,  multi15,  multipac, multipac, multipac,        ROT90,  "[Various] Clay Cowgill Hack", "Multipac (v1.3)" )
GAME( 1998, multi11,  multi15,  multipac, multipac, multipac,        ROT90,  "[Various] Clay Cowgill Hack", "Multipac (v1.1)" )
GAME( 1998, multi10,  multi15,  multipac, multipac, multipac,        ROT90,  "[Various] Clay Cowgill Hack", "Multipac (v1.0)" )
// End Multipac

// Pacman Super ABC (by Two-Bit Score)
// Works!

// Known Issues:
//   * ROM U14 is reported as bad in the diagnostics mode
//   * I haven't found a use for U18.ROM
//   * Can't find a dipswich to set the buy-in mode on the diagnostics screen

// Pac-Man Super ABC memory map (preliminary)
// (As Per Pacman with the following exceptions...
/*
0000-3fff ROM
4800-49ff RAM (High Score Stuff, I think)
8000-9fff ROM (Ms Pac-Man Games only)
a000-bfff ROM (Menu Routines used by all games)

memory mapped ports:

5006      coin lockout     // Top nyble holds which game is in play
& Bank Switching // (0 - pacman to 7 - Pacman Ultra)

ROM U14 is 8 banks of 0x10000 bytes, where in each 0x10000 byte region...
X0000-X3fff -> Memory addresses 0000-3fff
X4000-X5fff -> Memory addresses 8000-9fff
X6000-X9FFF -> Empty/Rubbish?
Xa000-Xbfff -> Memory addresses a000-bfff
Xc000-Xffff -> Empty/Rubbish?

*/

static void init_superabc(void)
{
}

INPUT_PORTS_START( superabc )
PORT_START	/* IN0 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
PORT_DIPSETTING(    0x00, DEF_STR( On ) )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

PORT_START	/* IN1 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
PORT_DIPNAME(0x80, 0x80, DEF_STR( Cabinet ) )
PORT_DIPSETTING(   0x80, DEF_STR( Upright ) )
PORT_DIPSETTING(   0x00, DEF_STR( Cocktail ) )

PORT_START	/* DSW 1 */
PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
PORT_DIPSETTING(    0x00, "1" )
PORT_DIPSETTING(    0x04, "2" )
PORT_DIPSETTING(    0x08, "3" )
PORT_DIPSETTING(    0x0c, "5" )
PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
PORT_DIPSETTING(    0x00, "10000" )
PORT_DIPSETTING(    0x10, "15000" )
PORT_DIPSETTING(    0x20, "20000" )
PORT_DIPSETTING(    0x30, "None" )
PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
PORT_DIPSETTING(    0x40, "Normal" )
PORT_DIPSETTING(    0x00, "Hard" )
PORT_DIPNAME( 0x80, 0x80, "Ghost Names" )
PORT_DIPSETTING(    0x80, "Normal" )
PORT_DIPSETTING(    0x00, "Alternate" )

PORT_START	/* DSW 2 */
PORT_DIPNAME( 0x01, 0x00, "DSW 2-1" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x01, "1" )
PORT_DIPNAME( 0x02, 0x00, "DSW 2-2" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x02, "1" )
PORT_DIPNAME( 0x04, 0x00, "DSW 2-3" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x04, "1" )
PORT_DIPNAME( 0x08, 0x00, "DSW 2-4" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x08, "1" )
PORT_DIPNAME( 0x10, 0x00, "DSW 2-5" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x10, "1" )
PORT_DIPNAME( 0x20, 0x00, "DSW 2-6" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x20, "1" )
PORT_DIPNAME( 0x40, 0x00, "DSW 2-7" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x40, "1" )
PORT_DIPNAME( 0x80, 0x00, "DSW 2-8" )
PORT_DIPSETTING(    0x00, "0" )
PORT_DIPSETTING(    0x80, "1" )
PORT_START	/* FAKE */
/* This fake input port is used to get the status of the fire button */
/* and activate the speedup cheat if it is. */
PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Speedup Cheat", KEYCODE_LCONTROL, JOYCODE_1_BUTTON1 )
PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
PORT_DIPSETTING(    0x01, DEF_STR( On ) )
INPUT_PORTS_END

static struct GfxDecodeInfo superabc_gfxdecodeinfo[] =
{
	{ REGION_GFX1, 0x00000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x01000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x10000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x11000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x08000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x09000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x18000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x19000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x04000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x05000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x14000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x15000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x0c000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x0d000, &spritelayout, 0, 32 },
	{ REGION_GFX1, 0x1c000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x1d000, &spritelayout, 0, 32 },                       
	{ -1 } /* end of array */
};

static MEMORY_READ_START( superabc_readmem )
{ 0x0000, 0x3fff, MRA_BANK1 },
{ 0x4000, 0x47ff, MRA_RAM },	/* video and color RAM */
{ 0x4800, 0x48ff, MRA_RAM },    // High Scores?
{ 0x4c00, 0x4fff, MRA_RAM },    /* including sprite codes at 4ff0-4fff */
{ 0x5000, 0x503f, input_port_0_r },	/* IN0 */
{ 0x5040, 0x507f, input_port_1_r },	/* IN1 */
{ 0x5080, 0x50bf, input_port_2_r },	/* DSW1 */
{ 0x50c0, 0x50ff, input_port_3_r },	/* DSW2 */
{ 0x8000, 0x9fff, MRA_BANK2 },          /* SuperABC */
{ 0xa000, 0xbfff, MRA_BANK3 },          /* SuperABC */
MEMORY_END

static MEMORY_WRITE_START( superabc_writemem )
{ 0x0000, 0x3fff, MWA_ROM },
{ 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
{ 0x4400, 0x47ff, colorram_w, &colorram },
{ 0x4800, 0x4900, MWA_RAM, &nvram, &nvram_size }, // High Scores?
{ 0x4c00, 0x4fef, MWA_RAM },
{ 0x4ff0, 0x4fff, MWA_RAM, &spriteram, &spriteram_size },
{ 0x5000, 0x5000, interrupt_enable_w },
{ 0x5001, 0x5001, pengo_sound_enable_w },
{ 0x5002, 0x5002, MWA_NOP },
{ 0x5003, 0x5003, pengo_flipscreen_w },
{ 0x5004, 0x5005, pacman_leds_w },
// 	{ 0x5006, 0x5006, pacman_coin_lockout_global_w },	this breaks many games
{ 0x5006, 0x5006, superabc_gfxbank_w }, 
{ 0x5007, 0x5007, pacman_coin_counter_w },
{ 0x5040, 0x505f, pengo_sound_w, &pengo_soundregs },
{ 0x5060, 0x506f, MWA_RAM, &spriteram_2 },
{ 0x50c0, 0x50c0, watchdog_reset_w },
{ 0x8000, 0xbfff, MWA_ROM },
{ 0xc000, 0xc3ff, videoram_w }, /* mirror address for video ram, */
{ 0xc400, 0xc7ef, colorram_w }, /* used to display HIGH SCORE and CREDITS */
{ 0xf000, 0xfffe, MWA_RAM },
{ 0xffff, 0xffff, MWA_NOP },    /* Eyes writes to this location to simplify code */
MEMORY_END

static MACHINE_DRIVER_START( superabc )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(superabc_readmem,superabc_writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(pacman_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(pacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(superabc_gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*64)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
//TODO
/* NV RAM Support */
//nvram_handler

MACHINE_DRIVER_END

ROM_START( superabc )
ROM_REGION( 0x80000, REGION_CPU1, 0 )   /* 8 banks of 64k for code */
ROM_LOAD( "u14.rom",      0x0000, 0x80000, 0xa560efe6 )

ROM_REGION( 0x20000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "u1.rom",       0x0000, 0x20000, 0x45caace0 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "7f.rom",       0x0000, 0x0020, 0x3a188666 )
ROM_LOAD( "4a.rom",       0x0020, 0x0100, 0x4382c049 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )  /* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

/*          rom       parent    machine   inp       init */
GAME( 1998, superabc, puckman,  superabc, superabc, superabc,        ROT90,  "[Various] Two-Bit Score Hack", "Pacman SuperABC" )

// End Pacman SuperABC

// Ms. Pacman II?

ROM_START( mspacii )
ROM_REGION( 0x10000, REGION_CPU1, 0 )   /* 8 banks of 64k for code */
ROM_LOAD( "msp2p36e.bin",  0x0000, 0x1000, 0xdf673b57 )
ROM_LOAD( "msp2p46f.bin",  0x1000, 0x1000, 0x7591f606 )
ROM_LOAD( "msp2p56h.bin",  0x2000, 0x1000, 0xc8ef1a7f )
ROM_LOAD( "msp2p66j.bin",  0x3000, 0x1000, 0x165a9dd8 )
ROM_LOAD( "msp2p7s1.bin",  0x8000, 0x1000, 0xfbbc3d2e )
ROM_LOAD( "msp2p8s2.bin",  0x9000, 0x1000, 0xaba3096d )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "msp2p15e.bin",  0x0000, 0x1000, 0x04333722 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "msp2p25f.bin",  0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",     0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",     0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )  
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

/*          rom       parent    machine   inp       init */
GAME( 1998, mspacii,  0,        pacman,   mspacman,   0,            ROT90,  "[Midway] Orca?", "Ms. Pacman II" )

/* Pac Gal (Uchida 82) driver */
/* Original Filename was PacGalJAMMA */

static struct GfxDecodeInfo pacgalu_gfxdecodeinfo[] =
{
	{ REGION_GFX1, 0x00000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x01000, &spritelayout, 0, 32 },
	{ -1 } /* end of array */
};

static MACHINE_DRIVER_START( pacgalu )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(readmem,writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(pacman_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(pacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(pacgalu_gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*64)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END

ROM_START( pacgalu )
ROM_REGION( 0x20000, REGION_CPU1, 0 )   
ROM_LOAD( "pgal1.bin", 0x10000, 0x8000, 0x23a0bd03 )
ROM_COPY( REGION_CPU1, 0x10000, 0x0000, 0x4000 )
ROM_COPY( REGION_CPU1, 0x14000, 0x8000, 0x2000 )

ROM_REGION( 0x2000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pgal2.bin",  0x0000, 0x2000, 0xd6e3a240 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	
ROM_END

/*          rom       parent    machine   inp       init */
GAME( 1998, pacgalu,  mspacman, pacgalu,  mspacman, 0,            ROT90,  "[Midway] Uchida (Bootleg)", "Pac-Gal (Uchida Bootleg)" )

/* Ms Pacman (Bootleg) */
/* Zip File mspacj */
/* This one is wierd. There are two idential copies of the first 16k of
program rom in mspac1.bin and four copies of the last 8k... all of which
are identical, so even if there were any bank switching there would be
no change...*/

static struct GfxDecodeInfo mspacj_gfxdecodeinfo[] =
{
	{ REGION_GFX1, 0x0000, &tilelayout,   0, 32 },
	{ REGION_GFX1, 0x1000, &spritelayout, 0, 32 },
	{ -1 } /* end of array */
};

static MACHINE_DRIVER_START( mspacj )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(readmem,writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(pacman_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(pacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(mspacj_gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*64)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END

ROM_START( mspacj )
ROM_REGION( 0x20000, REGION_CPU1, 0 )   
ROM_LOAD( "mspac1.bin",0x10000, 0x10000, 0x807f978a )
ROM_COPY( REGION_CPU1, 0x18000, 0x00000, 0x4000)
ROM_COPY( REGION_CPU1, 0x1c000, 0x08000, 0x2000)

ROM_REGION( 0x8000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "mspac2.bin", 0x4000, 0x4000, 0x8ee4a3b0 )

// Tiles
ROM_COPY( REGION_GFX1, 0x4000, 0x0000, 0x0800)
ROM_COPY( REGION_GFX1, 0x5000, 0x0800, 0x0800)

// Sprites
ROM_COPY( REGION_GFX1, 0x4800, 0x1000, 0x0800)
ROM_COPY( REGION_GFX1, 0x5800, 0x1800, 0x0800)


ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "mspacpr4.bin", 0x0000, 0x0020, 0x3545e7e9 )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "mspacpr2.bin", 0x0000, 0x0100, 0x640e455c )
ROM_LOAD( "mspacpr3.bin", 0x0100, 0x0100, 0x77245b66 )  
ROM_END

/*          rom       parent    machine   inp       init */
GAME( 1981, mspacj,   mspacman, mspacj,   mspacman, 0,        ROT90,  "bootleg", "Ms. Pac-Man (Jamma? Bootleg)" )

// Ms. Pacman/Galaga //
static struct GfxDecodeInfo pacgal20_gfxdecodeinfo[] =
{
	{ REGION_CPU1, 0x1a000, &tilelayout,   0, 32 }, // Ms. Pacman Tiles
	{ REGION_CPU1, 0x2d000, &tilelayout,   0, 32 }, // PacMan Tiles
	{ REGION_CPU1, 0x2a000, &tilelayout,   0, 32 }, // Galaga
	{ -1 } /* end of array */
};

static MACHINE_DRIVER_START( pacgal20 )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(readmem,writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(pacman_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(pacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(pacgal20_gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*64)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END

ROM_START( 20pacgal )
ROM_REGION( 0x50000, REGION_CPU1, 0 )   /* 64k for code + 256k for Banked ROM data */

ROM_LOAD( "20th_101"  ,0x10000, 0x40000, 0x77159582 )
ROM_COPY( REGION_CPU1, 0x2e000, 0x0000, 0x4000)

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	

ROM_END

/*          rom       parent    machine   inp       init */
GAME( 2000, 20pacgal, mspacman, pacgal20, mspacman, 0,        ROT90,  "Namco", "Class of '81: 20th Anniversary Ms.Pacman/Galaga" )

// Davids Ms. Pacman Bootleg Highscore MOD //

static MEMORY_WRITE_START( mspachi_writemem )
{ 0x0000, 0x3fff, MWA_ROM },
{ 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
{ 0x4400, 0x47ff, colorram_w, &colorram },
{ 0x4800, 0x49ff, MWA_RAM, &nvram, &nvram_size }, // High Score Table
{ 0x4c00, 0x4fef, MWA_RAM },
{ 0x4ff0, 0x4fff, MWA_RAM, &spriteram, &spriteram_size },
{ 0x5000, 0x5000, interrupt_enable_w },
{ 0x5001, 0x5001, pengo_sound_enable_w },
{ 0x5002, 0x5002, MWA_NOP },
{ 0x5003, 0x5003, pengo_flipscreen_w },
{ 0x5004, 0x5005, pacman_leds_w },
// 	{ 0x5006, 0x5006, pacman_coin_lockout_global_w },	this breaks many games
{ 0x5007, 0x5007, pacman_coin_counter_w },
{ 0x5040, 0x505f, pengo_sound_w, &pengo_soundregs },
{ 0x5060, 0x506f, MWA_RAM, &spriteram_2 },
{ 0x50c0, 0x50c0, watchdog_reset_w },
{ 0x8000, 0xbfff, MWA_ROM },	/* Ms. Pac-Man / Ponpoko only */
{ 0xc000, 0xc3ff, videoram_w }, /* mirror address for video ram, */
{ 0xc400, 0xc7ef, colorram_w }, /* used to display HIGH SCORE and CREDITS */
{ 0xffff, 0xffff, MWA_NOP },	/* Eyes writes to this location to simplify code */
MEMORY_END


static MACHINE_DRIVER_START( mspachi )

	/* basic machine hardware */
	MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
	MDRV_CPU_MEMORY(readmem,mspachi_writemem)
	MDRV_CPU_PORTS(0,writeport)
	MDRV_CPU_VBLANK_INT(pacman_interrupt,1)

	MDRV_FRAMES_PER_SECOND(60.606060)
	MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
	MDRV_MACHINE_INIT(pacman)

	/* video hardware */
	MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
	MDRV_SCREEN_SIZE(36*8, 28*8)
	MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
	MDRV_GFXDECODE(gfxdecodeinfo)
	MDRV_PALETTE_LENGTH(16)
	MDRV_COLORTABLE_LENGTH(4*32)

	MDRV_PALETTE_INIT(pacman)
	MDRV_VIDEO_START(pacman)
	MDRV_VIDEO_UPDATE(pengo)

	/* sound hardware */
	MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
//TODO
// NV RAM Support
//	nvram_handler

MACHINE_DRIVER_END

ROM_START( mspachi )
ROM_REGION( 0x11000, REGION_CPU1, 0 )	/* 64k for code */
ROM_LOAD( "boot1",        0x0000, 0x1000, 0xdca2ffc4 )
ROM_LOAD( "boot2",        0x1000, 0x1000, 0xfea2627b )
ROM_LOAD( "boot3",        0x2000, 0x1000, 0xa9a32749 )
ROM_LOAD( "boot4",        0x3000, 0x1000, 0x9123cd53 )
ROM_LOAD( "boot5",        0x8000, 0x1000, 0x397f196d )
ROM_LOAD( "boot6",        0x9000, 0x2000, 0x50de42d3 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "5e",           0x0000, 0x1000, 0x5c281d01 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "5f",           0x0000, 0x1000, 0x615af909 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

/*          rom       parent    machine   inp       init */
GAME( 1998, mspachi,  mspacman, mspachi,  mspacman, 0,        ROT90,  "[Midway] David Widel Kit", "Ms. Pacman (Dwidel's Hiscore Kit)" )

/*
Original, Encrypted Piranha set

This set is believed to be the "original" piranha set and uses
the same encryption as "eyes": It uses a slightly different CLUT prom (.4a)
than pacman, and the title screen is pretty much different to the piranha
set in MAME. Its easier to see how this set was taken for a new game
by arcade owners than the one existing in MAME

There is some protection which is not emulated, and is patched out in
init_newpir
*/

// This init is the same as for eyes, but with the code patch at 0x030fb
static void init_newpir(void)
{
	int i;
	unsigned char *RAM;

	/* CPU ROMs */

	/* Data lines D3 and D5 swapped */
	RAM = memory_region(REGION_CPU1);
	for (i = 0; i < 0x4000; i++)
	{
		RAM[i] =  (RAM[i] & 0xc0) | ((RAM[i] & 0x08) << 2) |
			(RAM[i] & 0x10) | ((RAM[i] & 0x20) >> 2) | (RAM[i] & 0x07);
	}

	/* Apply Code Patch to clear protection */
	RAM[0x30fb]=0xc3;
	RAM[0x30fc]=0x74;
	RAM[0x30fd]=0x31;

	/* Graphics ROMs */

	/* Data lines D4 and D6 and address lines A0 and A2 are swapped */
	RAM = memory_region(REGION_GFX1);
	for (i = 0;i < memory_region_length(REGION_GFX1);i += 8)
		eyes_decode(&RAM[i]);
	RAM = memory_region(REGION_GFX2);
	for (i = 0;i < memory_region_length(REGION_GFX2);i += 8)
		eyes_decode(&RAM[i]);
}

ROM_START( newpir )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
ROM_LOAD( "1.bin",        0x0000, 0x0800, 0x69a3e6ea )
ROM_LOAD( "5.bin",        0x0800, 0x0800, 0x245e753f )
ROM_LOAD( "2.bin",        0x1000, 0x0800, 0x62cb6954 )
ROM_LOAD( "6.bin",        0x1800, 0x0800, 0xcb0700bc )
ROM_LOAD( "3.bin",        0x2000, 0x0800, 0x843fbfe5 )
ROM_LOAD( "7.bin",        0x2800, 0x0800, 0x73084d5e )
ROM_LOAD( "4.bin",        0x3000, 0x0800, 0x4cdf6704 )
ROM_LOAD( "8.bin",        0x3800, 0x0800, 0xb86fedb3 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "9.bin",        0x0000, 0x0800, 0x0f19eb28 )
ROM_LOAD( "11.bin",       0x0800, 0x0800, 0x5f8bdabe )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )

ROM_LOAD( "10.bin",       0x0000, 0x0800, 0xd19399fb )
ROM_LOAD( "12.bin",       0x0800, 0x0800, 0xcfb4403d )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x08c9447b )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

ROM_START( newpirb )
ROM_REGION( 0x10000, REGION_CPU1, 0 )	/* 64k for code */
ROM_LOAD( "pir1.bin",        0x0000, 0x0800, 0xc6ce1bfc )
ROM_LOAD( "pir5.bin",        0x0800, 0x0800, 0xa2655a33 )
ROM_LOAD( "pir2.bin",        0x1000, 0x0800, 0x62cb6954 )
ROM_LOAD( "pir6.bin",        0x1800, 0x0800, 0xcb0700bc )
ROM_LOAD( "pir3.bin",        0x2000, 0x0800, 0x843fbfe5 )
ROM_LOAD( "pir7.bin",        0x2800, 0x0800, 0x73084d5e )
ROM_LOAD( "pir4.bin",        0x3000, 0x0800, 0x9363a4d1 )
ROM_LOAD( "pir8.bin",        0x3800, 0x0800, 0x2769979c )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "pir9.bin",        0x0000, 0x0800, 0x94eb7563 )
ROM_LOAD( "pir11.bin",       0x0800, 0x0800, 0xa3606973 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )

ROM_LOAD( "pir10.bin",       0x0000, 0x0800, 0x84165a2c )
ROM_LOAD( "pir12.bin",       0x0800, 0x0800, 0x2699ba9e )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "piranha.4a",    0x0020, 0x0100, 0x08c9447b )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

/*          rom       parent    machine   inp       init */
GAME( 1981, newpir,   puckman,  pacman,   mspacman, newpir,        ROT90,  "hack", "Piranha (Original - Encrypted)" )
GAME( 1981, newpirb,  puckman,  pacman,   mspacman, newpir,        ROT90,  "hack", "Piranha (Original - Encrypted/Older)" )

ROM_START( test )
ROM_REGION( 0x10000, REGION_CPU1, 0 )   /* 64k for code */
ROM_LOAD( "test.rom",     0x0000, 0x1000, 0xfb645998 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "test.5e",      0x0000, 0x1000, 0x0c944964 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "test.5f",      0x0000, 0x1000, 0x958fedf9 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )	/* sound PROMs */
ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )	/* timing - not used */
ROM_END

GAME( 1999, test,      puckman,   pacman,   mspacman, 0, ROT90, "David Caldwell", "Pacman Hardware Test" )

// Both NewPuc2 sets give bonus lives at different places then pacman
// The values don't tally with the self-test modes, but its the way the games
// work...

// New-Puck-2 Set 1
INPUT_PORTS_START( newpuc2 )
PORT_START	/* IN0 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
PORT_DIPSETTING(    0x00, DEF_STR( On ) )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

PORT_START	/* IN1 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
PORT_DIPNAME(0x80, 0x80, DEF_STR( Cabinet ) )
PORT_DIPSETTING(   0x80, DEF_STR( Upright ) )
PORT_DIPSETTING(   0x00, DEF_STR( Cocktail ) )

PORT_START	/* DSW 1 */
PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
PORT_DIPSETTING(    0x00, "1" )
PORT_DIPSETTING(    0x04, "2" )
PORT_DIPSETTING(    0x08, "3" )
PORT_DIPSETTING(    0x0c, "5" )
PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
PORT_DIPSETTING(    0x00, "20000" )
PORT_DIPSETTING(    0x10, "25000" )
PORT_DIPSETTING(    0x20, "30000" )
PORT_DIPSETTING(    0x30, "None" )
PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
PORT_DIPSETTING(    0x40, "Normal" )
PORT_DIPSETTING(    0x00, "Hard" )
PORT_DIPNAME( 0x80, 0x80, "Ghost Names" )
PORT_DIPSETTING(    0x80, "Normal" )
PORT_DIPSETTING(    0x00, "Alternate" )

PORT_START	/* DSW 2 */
PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )

PORT_START	/* FAKE */
/* This fake input port is used to get the status of the fire button */
/* and activate the speedup cheat if it is. */
PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Speedup Cheat", KEYCODE_LCONTROL, JOYCODE_1_BUTTON1 )
PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
PORT_DIPSETTING(    0x01, DEF_STR( On ) )
INPUT_PORTS_END

ROM_START( newpuc2 )
ROM_REGION( 0x10000, REGION_CPU1,0 )
ROM_LOAD( "6e.cpu",  0x0000, 0x0800, 0x69496a98 )
ROM_LOAD( "6k.cpu",  0x0800, 0x0800, 0x158fc01c )
ROM_LOAD( "6f.cpu",  0x1000, 0x0800, 0x7d177853 )
ROM_LOAD( "6m.cpu",  0x1800, 0x0800, 0x70810ccf )
ROM_LOAD( "6h.cpu",  0x2000, 0x0800, 0x81719de8 )
ROM_LOAD( "6n.cpu",  0x2800, 0x0800, 0x3f250c58 )
ROM_LOAD( "6j.cpu",  0x3000, 0x0800, 0xe6675736 )
ROM_LOAD( "6p.cpu",  0x3800, 0x0800, 0x1f81e765 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "5e.cpu",    0x0000, 0x0800, 0x2066a0b7 )
ROM_LOAD( "5h.cpu",    0x0800, 0x0800, 0x777c70d3 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "5f.cpu",    0x0000, 0x0800, 0xca8c184c )
ROM_LOAD( "5j.cpu",    0x0800, 0x0800, 0x7dc75a81 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f", 0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a", 0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )     /* sound PROMs */
ROM_LOAD( "82s126.1m", 0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m", 0x0100, 0x0100, 0x77245b66 )
ROM_END

/*          rom       parent     machine   inp       init */
GAME( 1981, newpuc2,  puckman,   pacman,   newpuc2,  0, ROT90, "[hack] Kamiya?", "New Puckman 2 (Set 1)" )

// New-Puck-2 Set 2
INPUT_PORTS_START( nwpuc2b )
PORT_START	/* IN0 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY )
PORT_BITX(    0x10, 0x10, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Rack Test", KEYCODE_F1, IP_JOY_NONE )
PORT_DIPSETTING(    0x10, DEF_STR( Off ) )
PORT_DIPSETTING(    0x00, DEF_STR( On ) )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_COIN1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_COIN2 )
PORT_BIT( 0x80, IP_ACTIVE_LOW, IPT_SERVICE1 )

PORT_START	/* IN1 */
PORT_BIT( 0x01, IP_ACTIVE_LOW, IPT_JOYSTICK_UP    | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x02, IP_ACTIVE_LOW, IPT_JOYSTICK_LEFT  | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x04, IP_ACTIVE_LOW, IPT_JOYSTICK_RIGHT | IPF_4WAY | IPF_COCKTAIL )
PORT_BIT( 0x08, IP_ACTIVE_LOW, IPT_JOYSTICK_DOWN  | IPF_4WAY | IPF_COCKTAIL )
PORT_SERVICE( 0x10, IP_ACTIVE_LOW )
PORT_BIT( 0x20, IP_ACTIVE_LOW, IPT_START1 )
PORT_BIT( 0x40, IP_ACTIVE_LOW, IPT_START2 )
PORT_DIPNAME(0x80, 0x80, DEF_STR( Cabinet ) )
PORT_DIPSETTING(   0x80, DEF_STR( Upright ) )
PORT_DIPSETTING(   0x00, DEF_STR( Cocktail ) )

PORT_START	/* DSW 1 */
PORT_DIPNAME( 0x03, 0x01, DEF_STR( Coinage ) )
PORT_DIPSETTING(    0x03, DEF_STR( 2C_1C ) )
PORT_DIPSETTING(    0x01, DEF_STR( 1C_1C ) )
PORT_DIPSETTING(    0x02, DEF_STR( 1C_2C ) )
PORT_DIPSETTING(    0x00, DEF_STR( Free_Play ) )
PORT_DIPNAME( 0x0c, 0x08, DEF_STR( Lives ) )
PORT_DIPSETTING(    0x00, "1" )
PORT_DIPSETTING(    0x04, "2" )
PORT_DIPSETTING(    0x08, "3" )
PORT_DIPSETTING(    0x0c, "5" )
PORT_DIPNAME( 0x30, 0x00, DEF_STR( Bonus_Life ) )
PORT_DIPSETTING(    0x00, "30000" )
PORT_DIPSETTING(    0x10, "60000" )
PORT_DIPSETTING(    0x20, "90000" )
PORT_DIPSETTING(    0x30, "None" )
PORT_DIPNAME( 0x40, 0x40, DEF_STR( Difficulty ) )
PORT_DIPSETTING(    0x40, "Normal" )
PORT_DIPSETTING(    0x00, "Hard" )
PORT_DIPNAME( 0x80, 0x80, "Ghost Names" )
PORT_DIPSETTING(    0x80, "Normal" )
PORT_DIPSETTING(    0x00, "Alternate" )

PORT_START	/* DSW 2 */
PORT_BIT( 0xff, IP_ACTIVE_HIGH, IPT_UNUSED )

PORT_START	/* FAKE */
/* This fake input port is used to get the status of the fire button */
/* and activate the speedup cheat if it is. */
PORT_BITX(    0x01, 0x00, IPT_DIPSWITCH_NAME | IPF_CHEAT, "Speedup Cheat", KEYCODE_LCONTROL, JOYCODE_1_BUTTON1 )
PORT_DIPSETTING(    0x00, DEF_STR( Off ) )
PORT_DIPSETTING(    0x01, DEF_STR( On ) )
INPUT_PORTS_END


ROM_START( nwpuc2b )
ROM_REGION( 0x10000, REGION_CPU1,0 )
ROM_LOAD( "nwpuc2b.6e", 0x0000, 0x1000, 0x532bd09f )
ROM_LOAD( "nwpuc2b.6f", 0x1000, 0x1000, 0xb9062f57 )
ROM_LOAD( "nwpuc2b.6h", 0x2000, 0x1000, 0x519b3c57 )
ROM_LOAD( "nwpuc2b.6j", 0x3000, 0x1000, 0x154cc118 )

ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
ROM_LOAD( "nwpuc2b.5e", 0x0000, 0x1000, 0x58357313 )

ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
ROM_LOAD( "nwpuc2b.5f", 0x0000, 0x1000, 0x4b044c87 )

ROM_REGION( 0x0120, REGION_PROMS, 0 )
ROM_LOAD( "82s123.7f", 0x0000, 0x0020, 0x2fc650bd )
ROM_LOAD( "82s126.4a", 0x0020, 0x0100, 0x3eb3a8e4 )

ROM_REGION( 0x0200, REGION_SOUND1, 0 )     /* sound PROMs */
ROM_LOAD( "82s126.1m", 0x0000, 0x0100, 0xa9cc86bf )
ROM_LOAD( "82s126.3m", 0x0100, 0x0100, 0x77245b66 )
ROM_END

/*          rom       parent     machine   inp       init */
GAME( 1981, nwpuc2b,  puckman,   pacman,   nwpuc2b,  0, ROT90, "[hack]", "New Puckman 2 (Set 2)" )

//Pacheart 2
static MEMORY_READ_START( pheartb_readmem )
 { 0x0000, 0x3f7f, MRA_ROM },
 { 0x3f80, 0x3fff, MRA_RAM },
 { 0x4000, 0x47ff, MRA_RAM },
 { 0x4c00, 0x4fff, MRA_RAM },
 { 0x5000, 0x503f, input_port_0_r },
 { 0x5040, 0x507f, input_port_1_r },
 { 0x5080, 0x50bf, input_port_2_r },
 { 0x50c0, 0x50ff, input_port_3_r },
MEMORY_END
 
static MEMORY_WRITE_START( pheartb_writemem )
 { 0x0000, 0x3f7f, MWA_ROM },
 { 0x3f80, 0x3fff, MWA_RAM },
 { 0x4000, 0x43ff, videoram_w, &videoram, &videoram_size },
 { 0x4400, 0x47ff, colorram_w, &colorram },
 { 0x4c00, 0x4fef, MWA_RAM },
 { 0x4ff0, 0x4fff, MWA_RAM, &spriteram, &spriteram_size },
 { 0x5000, 0x5000, interrupt_enable_w },
 { 0x5001, 0x5001, pengo_sound_enable_w },
 { 0x5002, 0x5002, MWA_NOP },
 { 0x5003, 0x5003, pengo_flipscreen_w },
  { 0x5004, 0x5005, pacman_leds_w },
//  { 0x5006, 0x5006, pacman_coin_lockout_global_w }, this breaks many games
  { 0x5007, 0x5007, pacman_coin_counter_w },
 { 0x5040, 0x505f, pengo_sound_w, &pengo_soundregs },
 { 0x5060, 0x506f, MWA_RAM, &spriteram_2 },
 { 0x50c0, 0x50c0, watchdog_reset_w },
 { 0xc000, 0xc3ff, videoram_w },
 { 0xc400, 0xc7ef, colorram_w },
MEMORY_END

static MACHINE_DRIVER_START( pheartb )
 
 MDRV_CPU_ADD_TAG("main", Z80, 18432000/6)
 MDRV_CPU_MEMORY(pheartb_readmem,pheartb_writemem)
 MDRV_CPU_PORTS(0,writeport)
 MDRV_CPU_VBLANK_INT(pacman_interrupt,1)
 
 MDRV_FRAMES_PER_SECOND(60.606060)
 MDRV_VBLANK_DURATION(DEFAULT_REAL_60HZ_VBLANK_DURATION)
 MDRV_MACHINE_INIT(pacman)
 
 MDRV_VIDEO_ATTRIBUTES(VIDEO_TYPE_RASTER)
 MDRV_SCREEN_SIZE(36*8, 28*8)
 MDRV_VISIBLE_AREA(0*8, 36*8-1, 0*8, 28*8-1)
 MDRV_GFXDECODE(gfxdecodeinfo)
 MDRV_PALETTE_LENGTH(16)
 MDRV_COLORTABLE_LENGTH(4*32)
 
 MDRV_PALETTE_INIT(pacman)
 MDRV_VIDEO_START(pacman)
 MDRV_VIDEO_UPDATE(pengo)
 
 MDRV_SOUND_ADD_TAG("namco", NAMCO, namco_interface)
MACHINE_DRIVER_END
 
ROM_START( pheartb )
 ROM_REGION( 0x10000, REGION_CPU1, 0 )
 ROM_LOAD( "1.bin",    0x0000, 0x0800, 0x69496a98)
 ROM_LOAD( "5.bin",    0x0800, 0x0800, 0x158fc01c )
 ROM_LOAD( "2.bin",    0x1000, 0x0800, 0x7d177853)
 ROM_LOAD( "6.bin",    0x1800, 0x0800, 0x70810ccf)
 ROM_LOAD( "3.bin",    0x2000, 0x0800, 0xb42416d1)
 ROM_LOAD( "7.bin",    0x2800, 0x0800, 0x3f250c58)
 ROM_LOAD( "4.bin",    0x3000, 0x0800, 0x01d13602)
 ROM_LOAD( "8.bin",    0x3800, 0x0800, 0x1f81e765)
 
 ROM_REGION( 0x1000, REGION_GFX1, ROMREGION_DISPOSE )
 ROM_LOAD( "9.bin",    0x0000, 0x0800, 0xc62bbabf )
 ROM_LOAD( "11.bin",   0x0800, 0x0800, 0x777c70d3 )
 
 ROM_REGION( 0x1000, REGION_GFX2, ROMREGION_DISPOSE )
 ROM_LOAD( "10.bin",   0x0000, 0x0800, 0xca8c184c )
 ROM_LOAD( "12.bin",   0x0800, 0x0800, 0x7dc75a81 )
 
 ROM_REGION( 0x0120, REGION_PROMS, 0 )
 ROM_LOAD( "82s123.7f",    0x0000, 0x0020, 0x2fc650bd )
 ROM_LOAD( "82s126.4a",    0x0020, 0x0100, 0x3eb3a8e4 )
 
 ROM_REGION( 0x0200, REGION_SOUND1, 0 )
 ROM_LOAD( "82s126.1m",    0x0000, 0x0100, 0xa9cc86bf )
 ROM_LOAD( "82s126.3m",    0x0100, 0x0100, 0x77245b66 )
ROM_END
 
GAME( 1981, pheartb,  puckman,  pheartb,  pacman,   0,        ROT90,  "Bootleg", "Pacman Hearts 2" )

